//==========================================================================
//
//      altera_bd_serial.c
//
//      Serial device driver for NIOS2 Altera_BD platform on-chip serial devices
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Hao Tseng base on ecos serial driver of MIPS IDT79s334a
//                                 and Altera ecos serial driver
// Contributors: 
// Date:         2006-02-28
// Purpose:      NIOS2 Altera_BD serial device driver
// Description:  
//
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <pkgconf/io_serial.h>
#include <pkgconf/io.h>

#include <cyg/io/io.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/io/devtab.h>
#include <cyg/infra/diag.h>
#include <cyg/io/serial.h>

#ifdef CYGPKG_IO_SERIAL_NIOS2_ALTERA_BD

#include CYGDAT_IO_SERIAL_DEVICE_HEADER
#include "altera_bd_serial.h"

typedef struct altera_bd_serial_info {
    CYG_ADDRWORD   base;
    CYG_WORD       int_num;
    cyg_interrupt  serial_interrupt;
    cyg_handle_t   serial_interrupt_handle;
} altera_bd_serial_info;

static bool altera_bd_serial_init(struct cyg_devtab_entry *tab);
static bool altera_bd_serial_putc(serial_channel *chan, unsigned char c);
static Cyg_ErrNo altera_bd_serial_lookup(struct cyg_devtab_entry **tab, 
                                   struct cyg_devtab_entry *sub_tab,
                                   const char *name);
static unsigned char altera_bd_serial_getc(serial_channel *chan);
static Cyg_ErrNo altera_bd_serial_set_config(serial_channel *chan, cyg_uint32 key,
                                         const void *xbuf, cyg_uint32 *len);
static void altera_bd_serial_start_xmit(serial_channel *chan);
static void altera_bd_serial_stop_xmit(serial_channel *chan);

static cyg_uint32 altera_bd_serial_ISR(cyg_vector_t vector, cyg_addrword_t data);
static void       altera_bd_serial_DSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);

static SERIAL_FUNS(altera_bd_serial_funs, 
                   altera_bd_serial_putc, 
                   altera_bd_serial_getc,
                   altera_bd_serial_set_config,
                   altera_bd_serial_start_xmit,
                   altera_bd_serial_stop_xmit
    );


#ifdef CYGPKG_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL0
static altera_bd_serial_info altera_bd_serial_info0 ={CYGHWR_NIOS2_UART0_ADDRESS,  
						  CYGHWR_NIOS2_UART0_VECTOR};

#if CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL0_BUFSIZE > 0
static unsigned char altera_bd_serial_out_buf0[CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL0_BUFSIZE];
static unsigned char altera_bd_serial_in_buf0[CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL0_BUFSIZE];

static SERIAL_CHANNEL_USING_INTERRUPTS(altera_bd_serial_channel0,
                                       altera_bd_serial_funs, 
                                       altera_bd_serial_info0,
                                       CYG_SERIAL_BAUD_RATE(CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL0_BAUD),
                                       CYG_SERIAL_STOP_DEFAULT,
                                       CYG_SERIAL_PARITY_DEFAULT,
                                       CYG_SERIAL_WORD_LENGTH_DEFAULT,
                                       CYG_SERIAL_FLAGS_DEFAULT,
                                       &altera_bd_serial_out_buf0[0], 
                                       sizeof(altera_bd_serial_out_buf0),
                                       &altera_bd_serial_in_buf0[0], 
                                       sizeof(altera_bd_serial_in_buf0)
    );
#else
static SERIAL_CHANNEL(altera_bd_serial_channel0,
                      altera_bd_serial_funs, 
                      altera_bd_serial_info0,
                      CYG_SERIAL_BAUD_RATE(CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL0_BAUD),
                      CYG_SERIAL_STOP_DEFAULT,
                      CYG_SERIAL_PARITY_DEFAULT,
                      CYG_SERIAL_WORD_LENGTH_DEFAULT,
                      CYG_SERIAL_FLAGS_DEFAULT
    );
#endif

DEVTAB_ENTRY(altera_bd_serial_io0, 
             CYGDAT_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL0_NAME,
             0,                 // Does not depend on a lower level interface
             &cyg_io_serial_devio, 
             altera_bd_serial_init, 
             altera_bd_serial_lookup,     // Serial driver may need initializing
             &altera_bd_serial_channel0
    );

#endif

#ifdef CYGPKG_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL1
static altera_bd_serial_info altera_bd_serial_info1 ={CYGHWR_NIOS2_UART1_ADDRESS,
						  CYGHWR_NIOS2_UART1_VECTOR};

#if CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL1_BUFSIZE > 0
static unsigned char altera_bd_serial_out_buf1[CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL1_BUFSIZE];
static unsigned char altera_bd_serial_in_buf1[CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL1_BUFSIZE];

static SERIAL_CHANNEL_USING_INTERRUPTS(altera_bd_serial_channel1,
                                       altera_bd_serial_funs,
                                       altera_bd_serial_info1,
                                       CYG_SERIAL_BAUD_RATE(CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL1_BAUD),
                                       CYG_SERIAL_STOP_DEFAULT,
                                       CYG_SERIAL_PARITY_DEFAULT,
                                       CYG_SERIAL_WORD_LENGTH_DEFAULT,
                                       CYG_SERIAL_FLAGS_DEFAULT,
                                       &altera_bd_serial_out_buf1[0],
                                       sizeof(altera_bd_serial_out_buf1),
                                       &altera_bd_serial_in_buf1[0],
                                       sizeof(altera_bd_serial_in_buf1)
    );
#else
static SERIAL_CHANNEL(altera_bd_serial_channel1,
                      altera_bd_serial_funs,
                      altera_bd_serial_info1,
                      CYG_SERIAL_BAUD_RATE(CYGNUM_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL1_BAUD),
                      CYG_SERIAL_STOP_DEFAULT,
                      CYG_SERIAL_PARITY_DEFAULT,
                      CYG_SERIAL_WORD_LENGTH_DEFAULT,
                      CYG_SERIAL_FLAGS_DEFAULT
    );
#endif

DEVTAB_ENTRY(altera_bd_serial_io1,
             CYGDAT_IO_SERIAL_NIOS2_ALTERA_BD_SERIAL1_NAME,
             0,                 // Does not depend on a lower level interface
             &cyg_io_serial_devio,
             altera_bd_serial_init,
             altera_bd_serial_lookup,     // Serial driver may need initializing
             &altera_bd_serial_channel1
    );

#endif

// Internal function to actually configure the hardware to desired baud rate, etc.
static bool
altera_bd_serial_config_port(serial_channel *chan, cyg_serial_info_t *new_config, bool init)
{
    altera_bd_serial_info *uart_chan = (altera_bd_serial_info *)chan->dev_priv;    
    cyg_uint32  port = uart_chan->base;
    cyg_uint16  baud_divisor;
    cyg_uint16  ctrl;

    /* 
     * There's no software control over the number of stop bits, parity or word 
     * length, so make sure it's not these that we're attempting to change.
     */

    if ((new_config->stop != chan->config.stop)     ||
        (new_config->parity != chan->config.parity) || 
       (new_config->word_length != chan->config.word_length)) 
    {
        return false;
    }
     
    // Disable port interrupts while changing hardware
    HAL_READ_UINT16(port+NIOS2_UART_CONTROL_REG_OFFSET, ctrl);
    HAL_WRITE_UINT16(port+NIOS2_UART_CONTROL_REG_OFFSET, 0);
    
    // clear all status 
    HAL_WRITE_UINT16(port+NIOS2_UART_STATUS_REG_OFFSET, 0);
    
    /*
     * Given that the request is good, update the hardware and the status in 
     * software.
     */
    baud_divisor=(cyg_uint16)((((cyg_uint32)CYGHWR_NIOS2_CPU_CLOCK_FREQ)/select_baud[new_config->baud]) - 1);
    HAL_WRITE_UINT16(port+NIOS2_UART_DIVISOR_REG_OFFSET, baud_divisor);
    
    if(init) {
        // enable RRDY interrupt
        ctrl |= UART_CTRL_BIT_IRRDY;
    }
    HAL_WRITE_UINT16(port+NIOS2_UART_CONTROL_REG_OFFSET, ctrl);
    
    if (new_config != &chan->config) {
        chan->config = *new_config;
    }
    return true;
}

// Function to initialize the device.  Called at bootstrap time.
static bool 
altera_bd_serial_init(struct cyg_devtab_entry *tab)
{
    serial_channel *chan = (serial_channel *)tab->priv;
//    altera_bd_serial_info *uart_chan = (altera_bd_serial_info *)chan->dev_priv;
#ifdef CYGDBG_IO_INIT
    diag_printf("Altera BD init - dev: %x.%d\n", uart_chan->base, uart_chan->int_num);
#endif
    (chan->callbacks->serial_init)(chan);  // Really only required for interrupt driven devices
//    if (chan->out_cbuf.len != 0) {
//        cyg_drv_interrupt_create(uart_chan->int_num,
//                                 0,         // can change IRQ0 priority
//                                 (cyg_addrword_t)chan,   //  Data item passed to interrupt handler
//                                 altera_bd_serial_ISR,
//                                 altera_bd_serial_DSR,
//                                 &uart_chan->serial_interrupt_handle,
//                                 &uart_chan->serial_interrupt);
//        cyg_drv_interrupt_attach(uart_chan->serial_interrupt_handle);
//        cyg_drv_interrupt_unmask(uart_chan->int_num);
//    }
//    altera_bd_serial_config_port(chan, &chan->config, true);
    return true;
}

// This routine is called when the device is "looked" up (i.e. attached)
static Cyg_ErrNo 
altera_bd_serial_lookup(struct cyg_devtab_entry **tab, 
                  struct cyg_devtab_entry *sub_tab,
                  const char *name)
{
    serial_channel *chan = (serial_channel *)(*tab)->priv;
    altera_bd_serial_info *uart_chan = (altera_bd_serial_info *)chan->dev_priv;
    
    (chan->callbacks->serial_init)(chan);  // Really only required for interrupt driven devices
    
    if (chan->out_cbuf.len != 0) {
        cyg_drv_interrupt_create(uart_chan->int_num,
                                 0,         // can change IRQ0 priority
                                 (cyg_addrword_t)chan,   //  Data item passed to interrupt handler
                                 altera_bd_serial_ISR,
                                 altera_bd_serial_DSR,
                                 &uart_chan->serial_interrupt_handle,
                                 &uart_chan->serial_interrupt);
        cyg_drv_interrupt_attach(uart_chan->serial_interrupt_handle);
        altera_bd_serial_config_port(chan, &chan->config, true);        
        cyg_drv_interrupt_unmask(uart_chan->int_num);
        return ENOERR;
    }
    else
        return EPERM;
    
}

// Send a character to the device output buffer.
// Return 'true' if character is sent to device
static bool
altera_bd_serial_putc(serial_channel *chan, unsigned char c)
{
    altera_bd_serial_info *uart_chan = (altera_bd_serial_info *)chan->dev_priv;    
    cyg_uint32  port = uart_chan->base;
    cyg_uint16  status;

    HAL_READ_UINT16(port+NIOS2_UART_STATUS_REG_OFFSET, status);
    if(status & UART_STATUS_BIT_TRDY) {
        // Transmit buffer is empty
        HAL_WRITE_UINT8(port+NIOS2_UART_TX_REG_OFFSET, c);
        return true;
    }else{
        // No space
        return false;
    }
}

// Fetch a character from the device input buffer, waiting if necessary
static unsigned char 
altera_bd_serial_getc(serial_channel *chan)
{
    unsigned char c;    
    altera_bd_serial_info *uart_chan = (altera_bd_serial_info *)chan->dev_priv;    
    cyg_uint32  port = uart_chan->base;
    cyg_uint16  status;

    do {
        HAL_READ_UINT16(port+NIOS2_UART_STATUS_REG_OFFSET, status);
    }while((status & UART_STATUS_BIT_RRDY) == 0);
    HAL_READ_UINT8(port+NIOS2_UART_RX_REG_OFFSET, c);
    return c;
}

// Set up the device characteristics; baud rate, etc.
static Cyg_ErrNo
altera_bd_serial_set_config(serial_channel *chan, cyg_uint32 key,
                        const void *xbuf, cyg_uint32 *len)
{
    switch (key) {
    case CYG_IO_SET_CONFIG_SERIAL_INFO:
      {
        cyg_serial_info_t *config = (cyg_serial_info_t *)xbuf;
        if ( *len < sizeof(cyg_serial_info_t) ) {
            return -EINVAL;
        }
        *len = sizeof(cyg_serial_info_t);
        if ( true != altera_bd_serial_config_port(chan, config, false) )
            return -EINVAL;
      }
      break;
    default:
        return -EINVAL;
    }
    return ENOERR;
}

// Enable the transmitter on the device
static void
altera_bd_serial_start_xmit(serial_channel *chan)
{
    altera_bd_serial_info *uart_chan = (altera_bd_serial_info *)chan->dev_priv;    
    cyg_uint32  port = uart_chan->base;
    cyg_uint16  ctrl;
    
    (chan->callbacks->xmt_char)(chan);
    HAL_READ_UINT16(port+NIOS2_UART_CONTROL_REG_OFFSET, ctrl);
    HAL_WRITE_UINT16(port+NIOS2_UART_CONTROL_REG_OFFSET, (ctrl | UART_CTRL_BIT_ITRDY));
}

// Disable the transmitter on the device
static void 
altera_bd_serial_stop_xmit(serial_channel *chan)
{
    altera_bd_serial_info *uart_chan = (altera_bd_serial_info *)chan->dev_priv;    
    cyg_uint32  port = uart_chan->base;
    cyg_uint16  ctrl;
    
    HAL_READ_UINT16(port+NIOS2_UART_CONTROL_REG_OFFSET, ctrl);
    HAL_WRITE_UINT16(port+NIOS2_UART_CONTROL_REG_OFFSET, (ctrl & (~UART_CTRL_BIT_ITRDY)));
    
}

// Serial I/O - low level interrupt handler (ISR)
static cyg_uint32 
altera_bd_serial_ISR(cyg_vector_t vector, cyg_addrword_t data)
{
    serial_channel *chan = (serial_channel *)data;
    altera_bd_serial_info *uart_chan = (altera_bd_serial_info *)chan->dev_priv;

    cyg_drv_interrupt_mask(uart_chan->int_num);
    cyg_drv_interrupt_acknowledge(uart_chan->int_num);
    return CYG_ISR_CALL_DSR;  // Cause DSR to be run
}

// Serial I/O - high level interrupt handler (DSR)
static void       
altera_bd_serial_DSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    serial_channel *chan = (serial_channel *)data;    
    altera_bd_serial_info *uart_chan = (altera_bd_serial_info *)chan->dev_priv;    
    cyg_uint32  port = uart_chan->base;
    cyg_uint16  status;
    cyg_uint8   cc;
    
    /*
     * Read the status register in order to determine the cause of the
     * interrupt.
     */

     HAL_READ_UINT16(port+NIOS2_UART_STATUS_REG_OFFSET, status);

     /* Clear any error flags set at the device */

     HAL_WRITE_UINT16(port+NIOS2_UART_STATUS_REG_OFFSET, 0);

     /* process a read irq */
 
     if (status & UART_STATUS_BIT_RRDY) {
         HAL_READ_UINT8(port+NIOS2_UART_RX_REG_OFFSET, cc);
         (chan->callbacks->rcv_char)(chan, cc);
     }
     
     if(status & UART_STATUS_BIT_TRDY) {
         (chan->callbacks->xmt_char)(chan);
     }
         
     
    cyg_drv_interrupt_unmask(uart_chan->int_num);
}

#endif

//-------------------------------------------------------------------------
// EOF altera_bd_serial.c
