#ifndef CYGONCE_HAL_CLOCK_H
	#define CYGONCE_HAL_CLOCK_H
//=============================================================================
//
//      hal_clock.h
//
//      HAL clock support
//
//=============================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):  
// Contributors: 
// Date:        
// Purpose:     Define clock support
// Description: The macros defined here provide the HAL APIs for handling
//              the clock.
//              
// Usage:
//              #include <cyg/hal/hal_intr.h> // which includes this file
//              ...
//              
//
//####DESCRIPTIONEND####
//
//=============================================================================
    #include <pkgconf/system.h>
	#include <cyg/hal/hal_io.h>	// nios II device IO register support
	#include CYGBLD_HAL_PLATFORM_H
   
// Timer Registers
enum
	{
	np_timerstatus_run_bit    = 1, // timer is running
	np_timerstatus_to_bit     = 0, // timer has timed out

	np_timercontrol_stop_bit  = 3, // stop the timer
	np_timercontrol_start_bit = 2, // start the timer
	np_timercontrol_cont_bit  = 1, // continous mode
	np_timercontrol_ito_bit   = 0, // enable time out interrupt

	np_timerstatus_run_mask    = (1<<1), // timer is running
	np_timerstatus_to_mask     = (1<<0), // timer has timed out

	np_timercontrol_stop_mask  = (1<<3), // stop the timer
	np_timercontrol_start_mask = (1<<2), // start the timer
	np_timercontrol_cont_mask  = (1<<1), // continous mode
	np_timercontrol_ito_mask   = (1<<0)  // enable time out interrupt
	};

    #define NIOS2_RTC_STATUS_REG      (CYGHWR_NIOS2_RTC_ADDRESS + 0x00 * 4)
    #define NIOS2_RTC_CONTROL_REG     (CYGHWR_NIOS2_RTC_ADDRESS + 0x01 * 4)
    #define NIOS2_RTC_PERIOD_L_REG    (CYGHWR_NIOS2_RTC_ADDRESS + 0x02 * 4)
    #define NIOS2_RTC_PERIOD_H_REG    (CYGHWR_NIOS2_RTC_ADDRESS + 0x03 * 4)
    #define NIOS2_RTC_SNAP_L_REG      (CYGHWR_NIOS2_RTC_ADDRESS + 0x04 * 4)
    #define NIOS2_RTC_SNAP_H_REG      (CYGHWR_NIOS2_RTC_ADDRESS + 0x05 * 4)    

	// initialise the clock preloaded value
externC void hal_clock_initialize(cyg_uint32 period);
externC void hal_clock_read(cyg_uint32 *pvalue);
#define HAL_CLOCK_INITIALIZE( _period_ )       hal_clock_initialize( _period_)
#define HAL_CLOCK_READ(_pvalue_)	           hal_clock_read( _pvalue_ )
#define HAL_CLOCK_RESET( _vector_, _period_ )

#ifdef CYGVAR_KERNEL_COUNTERS_CLOCK_LATENCY
	#define HAL_CLOCK_LATENCY( _pvalue_ )         HAL_CLOCK_READ( _pvalue_ )
#endif

//-----------------------------------------------------------------------------
#endif // ifndef CYGONCE_HAL_CLOCK_H
// End of hal_clock.h

