#ifndef CYGONCE_HAL_HAL_DIAG_H
#define CYGONCE_HAL_HAL_DIAG_H

/*=============================================================================
//
//      hal_diag.h
//
//      HAL Support for Kernel Diagnostic Routines
//
//=============================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Hao Tseng
// Contributors: 
// Date:         2003-12-15
// Purpose:      HAL Support for Kernel Diagnostic Routines
// Description:  Diagnostic routines for use during kernel development.
// Usage:        #include <cyg/hal/hal_diag.h>
//####DESCRIPTIONEND####
//========================================================================*/

#include <pkgconf/hal.h>
#include <pkgconf/system.h>
#include CYGBLD_HAL_PLATFORM_H

#include <cyg/infra/cyg_type.h>
#include <cyg/hal/hal_io.h>
#include <cyg/hal/hal_arch.h>

#if defined(CYGHWR_NIOS2_DIAG_USE_UART)
    #define NIOS2_DIAG_UART_RX_REG         (CYGHWR_NIOS2_DIAG_UART_ADDRESS + 0x00*4)
    #define NIOS2_DIAG_UART_TX_REG         (CYGHWR_NIOS2_DIAG_UART_ADDRESS + 0x01*4)
    #define NIOS2_DIAG_UART_STATUS_REG     (CYGHWR_NIOS2_DIAG_UART_ADDRESS + 0x02*4)
    #define NIOS2_DIAG_UART_CONTROL_REG    (CYGHWR_NIOS2_DIAG_UART_ADDRESS + 0x03*4)
    #define NIOS2_DIAG_UART_DIVISOR_REG    (CYGHWR_NIOS2_DIAG_UART_ADDRESS + 0x04*4)
    #define HAL_DIAG_LV_INIT()		HAL_WRITE_UINT16(NIOS2_DIAG_UART_CONTROL_REG, 0x00);
    #define HAL_DIAG_LV_WRITE_CHAR(_c_)									    \
    CYG_MACRO_START														    \
    {                                                                       \
        cyg_uint16 _read_=0;											    \
        HAL_READ_UINT16(NIOS2_DIAG_UART_STATUS_REG, _read_);		        \
        while(!(_read_ & 0x40))											    \
        HAL_READ_UINT16(NIOS2_DIAG_UART_STATUS_REG, _read_);		        \
        HAL_WRITE_UINT8(NIOS2_DIAG_UART_TX_REG, (_c_) );                    \
    }                                                                       \
    CYG_MACRO_END
#else
    #define HAL_DIAG_LV_INIT()
    #define HAL_DIAG_LV_WRITE_CHAR(_c_)
#endif


#if defined(CYGSEM_HAL_VIRTUAL_VECTOR_DIAG)

#include <cyg/hal/hal_if.h>

#define HAL_DIAG_INIT() hal_if_diag_init()
#define HAL_DIAG_WRITE_CHAR(_c_) hal_if_diag_write_char(_c_)
#define HAL_DIAG_READ_CHAR(_c_) hal_if_diag_read_char(&_c_)


#else // everything by steam

   #if defined(CYGHWR_NIOS2_DIAG_USE_UART)  
	   // initialise the UART
	   // disable all interrupt from UART
	   #define HAL_DIAG_INIT()		HAL_WRITE_UINT16(NIOS2_DIAG_UART_CONTROL_REG, 0x00);
		
	   // macro write a character to UART1
	   #define HAL_DIAG_WRITE_CHAR(_c_)										  \
	   CYG_MACRO_START														  \
	   cyg_uint16 _read_=0;													  \
	   if( 1|| (_c_) != '\r' )												  \
	   {																	  \
           HAL_READ_UINT16(NIOS2_DIAG_UART_STATUS_REG, _read_);				  \
           while(!(_read_ & 0x40))											  \
               HAL_READ_UINT16(NIOS2_DIAG_UART_STATUS_REG, _read_);			  \
           HAL_WRITE_UINT8(NIOS2_DIAG_UART_TX_REG, (_c_) );                    \
       }																	  \
       CYG_MACRO_END
       
       #define HAL_DIAG_READ_CHAR(_c_) (_c_) = 0;       
       
   
   #else
       #define HAL_DIAG_INIT() 
       #define HAL_DIAG_WRITE_CHAR(_c_) 
       #define HAL_DIAG_READ_CHAR(_c_) (_c_) = 0;
   #endif

#endif // CYGSEM_HAL_VIRTUAL_VECTOR_DIAG


#define _DUMP_LV_BYTE(_cc_)                                 \
CYG_MACRO_START											    \
{                                                           \
    if((_cc_ >> 4) >= 10)                                   \
        HAL_DIAG_LV_WRITE_CHAR('a' + ((_cc_) >> 4) - 10);   \
    else                                                    \
        HAL_DIAG_LV_WRITE_CHAR('0' + ((_cc_) >> 4));        \
    if((_cc_ & 0x0f) >= 10)                                 \
        HAL_DIAG_LV_WRITE_CHAR('a' + ((_cc_) & 0x0f) - 10); \
    else                                                    \
        HAL_DIAG_LV_WRITE_CHAR('0' + ((_cc_) & 0x0f));      \
}                                                           \
CYG_MACRO_END

#define _DUMP_LV_UINT16(_uint16_)                           \
CYG_MACRO_START											    \
{                                                           \
    _DUMP_LV_BYTE(((_uint16_) & 0xff00)>>8);                \
    _DUMP_LV_BYTE(((_uint16_) & 0xff));                     \
}                                                           \
CYG_MACRO_END

#define _DUMP_LV_UINT32(_uint32_)                           \
CYG_MACRO_START											    \
{                                                           \
    _DUMP_LV_UINT16(((_uint32_) & 0xffff)>>16);             \
    _DUMP_LV_UINT16(((_uint32_) & 0xffff));                 \
}                                                           \
CYG_MACRO_END

#define _GET_CPU_CURRENT_CWP(_cwp_)                         \
CYG_MACRO_START                                             \
{                                                           \
	asm volatile (                                          \
		"rdctl %0 \n					                    \
		pfx  %%hi(0x1f0) \n					                \
		and  %0, %%lo(0x1f0) \n		                        \
		lsri %0, 4 "                                        \
		: "=r"(_cwp_)                                       \
		);                                                  \
}                                                           \
CYG_MACRO_END
    

externC void _DUMP_BYTE(unsigned char cc);
externC void _DUMP_UINT16(cyg_uint16 uu);
externC void _DUMP_UINT32(cyg_uint32 uu);
externC void _DUMP_CHAR(unsigned char cc);
externC void _DUMP_STRING(unsigned char *str);
externC void _DUMP_RETURN(void);
externC void _DUMP_REGS(HAL_SavedRegisters *Regs);
externC void _DUMP_MEM(cyg_uint32 addr, cyg_uint32 len);

#endif /* CYGONCE_HAL_DIAG_H */
/*---------------------------------------------------------------------------*/
/* end of hal_diag.h                                                         */

