##=============================================================================
##
##	platform.s
##
##	NIOS II altera cyclone platform code
##
##=============================================================================
#####ECOSGPLCOPYRIGHTBEGIN####
## -------------------------------------------
## This file is part of eCos, the Embedded Configurable Operating System.
## Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
##
## eCos is free software; you can redistribute it and/or modify it under
## the terms of the GNU General Public License as published by the Free
## Software Foundation; either version 2 or (at your option) any later version.
##
## eCos is distributed in the hope that it will be useful, but WITHOUT ANY
## WARRANTY; without even the implied warranty of MERCHANTABILITY or
## FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
## for more details.
##
## You should have received a copy of the GNU General Public License along
## with eCos; if not, write to the Free Software Foundation, Inc.,
## 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## As a special exception, if other files instantiate templates or use macros
## or inline functions from this file, or you compile this file and link it
## with other works to produce a work based on this file, this file does not
## by itself cause the resulting work to be covered by the GNU General Public
## License. However the source code for this file must still be made available
## in accordance with section (3) of the GNU General Public License.
##
## This exception does not invalidate any other reasons why a work based on
## this file might be covered by the GNU General Public License.
##
## Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
## at http://sources.redhat.com/ecos/ecos-license/
## -------------------------------------------
#####ECOSGPLCOPYRIGHTEND####
##=============================================================================
#######DESCRIPTIONBEGIN####
##
## Author(s): 	 Hao Tseng
## Contributors: 
## Date:	 2004/04/05
## Purpose:	
## Description:	 NIOS II altera cyclone platform code
##              
######DESCRIPTIONEND####
##
##=============================================================================

#include <pkgconf/system.h>	
#include <pkgconf/hal.h>
#include <cyg/hal/arch.inc>
#include <cyg/hal/nios2_imacros.inc>
#include CYGBLD_HAL_PLATFORM_H

//-----------------------------------------------------------------------------
// (1) Platform Reset
//
    .text
    .globl hal_platform_reset
hal_platform_reset:

    // Initial basic hardware, such as DRAM
    
    
  #if CYGHWR_NIOS2_DIAG_USE_UART
    // Disable diagnostic UART's interrupt
    movhi   r1, %hi(CYGHWR_NIOS2_DIAG_UART_ADDRESS)
    ori     r1,r1,%lo(CYGHWR_NIOS2_DIAG_UART_ADDRESS)
    sthio   r0, 12(r1)
  #endif    

    // Disable RTC's interrupt
    movhi   r1, %hi(CYGHWR_NIOS2_RTC_ADDRESS)
    ori     r1,r1,%lo(CYGHWR_NIOS2_RTC_ADDRESS)
    addi    r2,r0,8
    sthio   r2, 4(r1)

    // clear RTC flag    
    sthio   r0, 0(r1)    

    ret
//-----------------------------------------------------------------------------
// (2) Interrupt pre_process
//

    .text
    .globl hal_platform_intr_pre_process
hal_platform_intr_pre_process:
    // r4 = irq_number

    // Clear peripheral's interrupt
    
    //Is this Interrupt for RTC ?
    addi    r1,r4,-CYGHWR_NIOS2_RTC_VECTOR
    bne     r1,r0, _intr_pre_proc_exit
    
_clear_RTC_intr:
    // clear RTC flag
    movhi   r1, %hi(CYGHWR_NIOS2_RTC_ADDRESS)
    ori     r1,r1,%lo(CYGHWR_NIOS2_RTC_ADDRESS)
    sthio   r0, 0(r1)    

_intr_pre_proc_exit:    
    ret

//-----------------------------------------------------------------------------
// (3) Cache process
#ifdef CYGHWR_NIOS2_HAS_ICACHE
    .globl hal_flush_icache
hal_flush_icache:
	addi	sp, sp, -12
	stw	fp, 8(sp)
	mov	fp, sp
	stw	r4, 0(fp)
	stw	r5, 4(fp)

    // flush icache    
    mov     r4, r0
    movhi   r5, %hi(CYGHWR_NIOS2_ICACHE_SIZE)
    ori     r5, r5, %lo(CYGHWR_NIOS2_ICACHE_SIZE)
__icache_flush_loop:
    flushi  r4
    addi    r4, r4, CYGHWR_NIOS2_ICACHE_LINE_SIZE
    flushi  r4
    addi    r4, r4, CYGHWR_NIOS2_ICACHE_LINE_SIZE
    bltu    r4, r5, __icache_flush_loop
    flushp
    
    ldw r4, 0(fp)
    ldw r5, 4(fp)    
	ldw	fp, 8(sp)
	addi	sp, sp, 12
	ret	
#endif

#ifdef CYGHWR_NIOS2_HAS_DCACHE
    .globl hal_flush_dcache
hal_flush_dcache:
	addi	sp, sp, -16
	stw	fp, 12(sp)
	mov	fp, sp
	stw	r4, 0(fp)
	stw	r5, 4(fp)

    //Flush all dcache
    mov     r4,r0
    movhi   r5, %hi(CYGHWR_NIOS2_DCACHE_SIZE)
    ori     r5, r5, %lo(CYGHWR_NIOS2_DCACHE_SIZE)
__dcache_flush_loop:
    flushd  0*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    flushd  1*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    flushd  2*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    flushd  3*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    flushd  4*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    flushd  5*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    flushd  6*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    flushd  7*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    addi    r4, r4, CYGHWR_NIOS2_DCACHE_LINE_SIZE * 8
    bltu    r4, r5, __dcache_flush_loop    
    
    ldw r4, 0(fp)
    ldw r5, 4(fp)
	ldw	fp, 12(sp)
	addi	sp, sp, 16
	ret	
#endif




    
    
##-----------------------------------------------------------------------------
## end of platform.S
	
