//===========================================================================
//
//      plf_intr.c
//
//      NIOS II altera cyclone board interrupt control
//
//===========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//===========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Hao Tseng
// Contributors: 
// Date:         2004/01/12
// Purpose:      interrupt control for NIOS II altera cyclone board.
//              
//####DESCRIPTIONEND####
//
//===========================================================================

#include <pkgconf/system.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/hal/hal_io.h>
#include <cyg/hal/hal_clock.h>
#include <cyg/hal/hal_diag.h>
#include CYGBLD_HAL_PLATFORM_H

// RTC status 
static cyg_uint16   RTC_start=0;

void hal_interrupt_mask(int _vector_)
{
    cyg_uint16  rdata;
    cyg_uint32 ie_reg;

    HAL_GET_IENABLE_REG(ie_reg);
    ie_reg &= (~(1 << _vector_));
    HAL_SET_IENABLE_REG(ie_reg);
    
    switch(_vector_)
    {
        case CYGNUM_HAL_INTERRUPT_RTC:
            // Disable RTC interrupt 
            HAL_READ_UINT16(NIOS2_RTC_CONTROL_REG,rdata);
            HAL_WRITE_UINT16(NIOS2_RTC_CONTROL_REG,(rdata & 0xfffe));
            break;
#ifdef CYGHWR_NIOS2_DIAG_USE_UART
        case CYGNUM_HAL_INTERRUPT_DIAG_UART:
            // Disable DIAG_UART interrupt 
            HAL_READ_UINT16(NIOS2_DIAG_UART_CONTROL_REG,rdata);
            HAL_WRITE_UINT16(NIOS2_DIAG_UART_CONTROL_REG,(rdata & 0xfeff));            
            break;
#endif
    }
    
}

void hal_interrupt_unmask(int _vector_)
{
    cyg_uint16  rdata;
    cyg_uint32 ie_reg;

    HAL_GET_IENABLE_REG(ie_reg);
    ie_reg |= (1 << _vector_);
    HAL_SET_IENABLE_REG(ie_reg);
    
     switch(_vector_)
    {
        case CYGNUM_HAL_INTERRUPT_RTC:
            // Enable RTC interrupt 
            HAL_READ_UINT16(NIOS2_RTC_CONTROL_REG,rdata);
            HAL_WRITE_UINT16(NIOS2_RTC_CONTROL_REG,(rdata | 0x1));    
            break;
#ifdef CYGHWR_NIOS2_DIAG_USE_UART
        case CYGNUM_HAL_INTERRUPT_DIAG_UART:
            // Enable DIAG_UART interrupt 
            HAL_READ_UINT16(NIOS2_DIAG_UART_CONTROL_REG,rdata);
            HAL_WRITE_UINT16(NIOS2_DIAG_UART_CONTROL_REG,(rdata & 0x100));            
            break;
#endif
    }
 
}

void hal_interrupt_acknowledge(int _vector_)
{
    // do nothing.
}

// RTC control
void hal_clock_initialize(cyg_uint32 period)
{
     HAL_WRITE_UINT16(NIOS2_RTC_PERIOD_L_REG,(cyg_uint16)(period & 0x0ffff));					     
     HAL_WRITE_UINT16(NIOS2_RTC_PERIOD_H_REG,(cyg_uint16)((period >> 16)));			                 
     HAL_WRITE_UINT16(NIOS2_RTC_CONTROL_REG,                                                           
                 (np_timercontrol_start_mask | np_timercontrol_cont_mask));
     RTC_start = 1;
}

void hal_clock_read(cyg_uint32 *pvalue)
{
    cyg_uint16 temp;                                                 
    cyg_uint32 period;              
    // If RTC did not initialized, then init it first
    if(RTC_start == 0)
        hal_clock_initialize(CYGNUM_HAL_RTC_PERIOD);
    
    HAL_READ_UINT16(NIOS2_RTC_PERIOD_L_REG, temp);                             
    HAL_READ_UINT16(NIOS2_RTC_PERIOD_H_REG, period);                           
    period = (period << 16) | (temp & 0xffff);                                
    HAL_WRITE_UINT16(NIOS2_RTC_SNAP_L_REG, 0);                                 
    HAL_READ_UINT16(NIOS2_RTC_SNAP_L_REG, temp);                               
    HAL_READ_UINT16(NIOS2_RTC_SNAP_H_REG, *pvalue);        
    *pvalue = ((*pvalue) << 16) | (temp & 0xffff); 
    *pvalue = period - (*pvalue);	      
}
