#ifndef CYGONCE_HAL_HAL_INTR_H
#define CYGONCE_HAL_HAL_INTR_H

//==========================================================================
//
//      hal_intr.h
//
//      HAL Interrupt and clock support
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//===========================================================================
//
// Author(s):    Hao Tseng
// Contributors: None
//               
// Date:         2003-11-12
// Purpose:      Define Interrupt support
// Description:  The macros defined here provide the HAL APIs for handling
//               interrupts and the clock.
//              
// Usage:
//               #include <cyg/hal/hal_intr.h>
//               ...
//              
//
//####DESCRIPTIONEND####


#include <pkgconf/hal.h>

#include <cyg/infra/cyg_type.h>
#include <cyg/hal/hal_io.h>

//--------------------------------------------------------------------------
// (1) Execption vectors definition
// The vectors definition is put at arch.inc
#include <cyg/hal/arch.inc>
#include <cyg/hal/plf_intr.h>

//--------------------------------------------------------------------------
// (2) Map any hardware execption to standard names


//--------------------------------------------------------------------------
// (3) Static data to handle interrupts and execptions

// ISR tables
externC volatile CYG_ADDRESS    hal_interrupt_handlers[CYGNUM_HAL_ISR_COUNT];
externC volatile CYG_ADDRWORD   hal_interrupt_data[CYGNUM_HAL_ISR_COUNT];
externC volatile CYG_ADDRESS    hal_interrupt_objects[CYGNUM_HAL_ISR_COUNT];

externC volatile CYG_ADDRESS    hal_vsr_table[CYGNUM_HAL_EXCEPTION_COUNT];

//--------------------------------------------------------------------------
// Default ISR
// The #define is used to test whether this routine exists, and to allow
// us to call it.
externC cyg_uint32 hal_default_isr(CYG_ADDRWORD vector, CYG_ADDRWORD data);
#define HAL_DEFAULT_ISR hal_default_isr

//--------------------------------------------------------------------------
// (4) Interrupt control macros

// Interrupt state storage
typedef cyg_uint32 CYG_INTERRUPT_STATE;

// set PIE = 0x0, return status reg
#define HAL_DISABLE_INTERRUPTS(_old_)	\
{                                       \
	asm volatile (                      \
        "rdctl %0, status \n            \
        wrctl status,r0 \n"             \
        : "=r"(_old_)                   \
        );                              \
}

// set PIE = 1
#define HAL_ENABLE_INTERRUPTS()		    \
{                                       \
	asm volatile (                      \
        "addi  at,r0,1 \n               \
        wrctl  status,at \n"            \
		:                               \
		:                               \
		: "at"                          \
		);                              \
}        

// set status = _old_
#define HAL_RESTORE_INTERRUPTS(_old_)	\
{                                       \
	asm volatile (                      \
        "wrctl status, %0\n"            \
        :                               \
        : "r"(_old_)                    \
        );                              \
}        

// return status
#define HAL_QUERY_INTERRUPTS( _state_ )         \
{                                               \
    asm volatile (                              \
        "rdctl %0, status\n"                    \
        : "=r"(_state_)                         \
        );                                      \
}

#define HAL_GET_IENABLE_REG( _value_ )  \
{                                       \
	asm volatile (                      \
        "rdctl %0, ienable \n"          \
        : "=r"(_value_)                 \
        );                              \
}

#define HAL_SET_IENABLE_REG(_value_)	\
{                                       \
	asm volatile (                      \
        "wrctl ienable, %0\n"           \
        :                               \
        : "r"(_value_)                  \
        );                              \
}        


//--------------------------------------------------------------------------
// (5 )Routine to execute DSRs using separate interrupt stack
#ifdef  CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
externC void hal_interrupt_stack_call_pending_DSRs(void);
#define HAL_INTERRUPT_STACK_CALL_PENDING_DSRS() \
    hal_interrupt_stack_call_pending_DSRs()

// these are offered solely for stack usage testing
// if they are not defined, then there is no interrupt stack.
#define HAL_INTERRUPT_STACK_BASE cyg_interrupt_stack_base
#define HAL_INTERRUPT_STACK_TOP  cyg_interrupt_stack
// use them to declare these extern however you want:
//       extern char HAL_INTERRUPT_STACK_BASE[];
//       extern char HAL_INTERRUPT_STACK_TOP[];
// is recommended
#endif

//--------------------------------------------------------------------------
// (6) Interrupt and VSR attachment macros

// Vector translation.
// For chained interrupts we only have a single vector though which all
// are passed. For unchained interrupts we have a vector per interrupt.

#ifndef HAL_TRANSLATE_VECTOR
#define HAL_TRANSLATE_VECTOR(_vector_,_index_) (_index_) = (_vector_)
#endif

#define HAL_INTERRUPT_IN_USE( _vector_, _state_)                             \
	CYG_MACRO_START                                                          \
	cyg_uint32 _index_;                                                      \
	HAL_TRANSLATE_VECTOR ((_vector_), _index_);                              \
																			 \
	if( (CYG_ADDRESS)hal_default_isr  == hal_interrupt_handlers[_vector_])	 \
    {																	     \
		(_state_) = 0;                                                       \
	} else {                                                                 \
		(_state_) = 1;                                                       \
	}                                                                        \
	CYG_MACRO_END

#define HAL_INTERRUPT_ATTACH( _vector_, _isr_, _data_, _object_ )           \
	CYG_MACRO_START                                                         \
	if( (CYG_ADDRESS)hal_default_isr  == hal_interrupt_handlers[_vector_])  \
	{																		\
		hal_interrupt_handlers[_vector_] = (CYG_ADDRESS)_isr_;              \
		hal_interrupt_data[_vector_] = (CYG_ADDRWORD) _data_;               \
		hal_interrupt_objects[_vector_] = (CYG_ADDRESS)_object_;            \
	}                                                                       \
CYG_MACRO_END                                                           
                                                                            
#define HAL_INTERRUPT_DETACH( _vector_, _isr_ ) CYG_MACRO_START             \
	if( hal_interrupt_handlers[_vector_] == (CYG_ADDRESS)_isr_ )            \
	{                                                                       \
		hal_interrupt_handlers[_vector_] = (CYG_ADDRESS)hal_default_isr;    \
		hal_interrupt_data[_vector_] = 0;                                   \
		hal_interrupt_objects[_vector_] = 0;                                \
	}                                                                       \
CYG_MACRO_END

#define HAL_VSR_GET( _vector_, _pvsr_ )                                     \
	*(CYG_ADDRESS *)(_pvsr_) = hal_vsr_table[_vector_];
    

#define HAL_VSR_SET( _vector_, _vsr_, _poldvsr_ ) CYG_MACRO_START           \
	if( _poldvsr_ != NULL )                                                 \
		*(CYG_ADDRESS *)_poldvsr_ = hal_vsr_table[_vector_];                \
	hal_vsr_table[_vector_] = (CYG_ADDRESS)_vsr_;                           \
CYG_MACRO_END

// This is an ugly name, but what it means is: grab the VSR back to eCos
// internal handling, or if you like, the default handler.  But if
// cooperating with GDB and CygMon, the default behaviour is to pass most
// exceptions to CygMon.  This macro undoes that so that eCos handles the
// exception.  So use it with care.

#define HAL_VSR_SET_TO_ECOS_HANDLER( _vector_, _poldvsr_ ) CYG_MACRO_START  \
	if( _poldvsr_ != NULL )                                                 \
		*(CYG_ADDRESS *)_poldvsr_ = hal_vsr_table[_vector_];                \
	hal_vsr_table[_vector_] = (CYG_ADDRESS)hal_default_interrupt_vsr;       \
CYG_MACRO_END


//--------------------------------------------------------------------------
// Interrupt controller access

#ifndef CYGHWR_HAL_INTERRUPT_CONTROLLER_ACCESS_DEFINED
externC void hal_interrupt_mask(int _vector_);
externC void hal_interrupt_unmask(int _vector_);
externC void hal_interrupt_acknowledge(int _vector_);
#define HAL_INTERRUPT_MASK( _vector_ )      hal_interrupt_mask(_vector_)
#define HAL_INTERRUPT_UNMASK( _vector_ )    hal_interrupt_unmask(_vector_)
#define HAL_INTERRUPT_ACKNOWLEDGE( _vector_ )   hal_interrupt_acknowledge(_vector_)
#define HAL_INTERRUPT_CONFIGURE( _vector_, _level_, _up_ )
#define HAL_INTERRUPT_SET_LEVEL( _vector_, _level_ )

#define CYGHWR_HAL_INTERRUPT_CONTROLLER_ACCESS_DEFINED

#endif


//--------------------------------------------------------------------------
// Clock control.
#include <cyg/hal/hal_clock.h>


//--------------------------------------------------------------------------
// Microsecond delay function provided in hal_misc.c
externC void hal_delay_us(int us);
#define HAL_DELAY_US(n)          hal_delay_us(n)


#endif // ifndef CYGONCE_HAL_HAL_INTR_H
// End of hal_intr.h

