//=============================================================================
//	context.S
//
//	NIOS II context switch code
//
//=============================================================================
//###ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//###ECOSGPLCOPYRIGHTEND####
//=============================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s): 	Hao Tseng
// Contributors:
// Date:	
// Purpose:	NIOS II context switch code
// Description:	This file contains implementations of the thread context 
//		        switch routines. It also contains the longjmp() and setjmp()
//              routines.
//
//####DESCRIPTIONEND####
//
//=============================================================================

#include <pkgconf/hal.h>
#include <cyg/hal/arch.inc>
#include <cyg/hal/nios2_imacros.inc>

#ifdef at
#undef at
#endif

// hal_thread_switch_context
// switch thread contexts
// r4 = address of sp of next thread to excute
// r5 = address of sp save location of current thread

	.global hal_thread_switch_context
	// .global makes the symbol visible to ld.
hal_thread_switch_context:
	
	// the parameter passed back (location store the stack pointer)	
	// Note: r5 passed from caller is the address of where the old thread pointer to be stored
    stw     sp, 0(r5)

    addi    sp,sp, -SAVED_REGS_SIZE
    stw     r0, 0(sp)
    stw     r1, 4(sp)
    stw     r2, 8(sp)
    stw     r3, 12(sp)
    stw     r4, 16(sp)
    stw     r5, 20(sp)
    stw     r6, 24(sp)
    stw     r7, 28(sp)
    stw     r8, 32(sp)
    stw     r9, 36(sp)
    stw     r10, 40(sp)
    stw     r11, 44(sp)
    stw     r12, 48(sp)
    stw     r13, 52(sp)
    stw     r14, 56(sp)
    stw     r15, 60(sp)
    stw     r16, 64(sp)
    stw     r17, 68(sp)
    stw     r18, 72(sp)
    stw     r19, 76(sp)
    stw     r20, 80(sp)
    stw     r21, 84(sp)
    stw     r22, 88(sp)
    stw     r23, 92(sp)
    stw     r24, 96(sp)
    stw     r25, 100(sp)
    stw     gp, 104(sp)
    addi    r1,sp,SAVED_REGS_SIZE
    stw     r1, 108(sp)
    stw     fp, 112(sp)
    stw     ea, 116(sp)
    stw     ba, 120(sp)
    stw     ra, 124(sp)
    // save ctrl registers
    rdctl   r1, status
    stw     r1, 128(sp)
    rdctl   r1, estatus
    stw     r1, 132(sp)
    rdctl   r1, bstatus
    stw     r1, 136(sp)
    rdctl   r1, ienable
    stw     r1, 140(sp)
    rdctl   r1, ipending
    stw     r1, 144(sp)

 // all "*old*" thread's register windows have been saved 
// now we can restore the "*new*" thread's register window
// by droping through into "hal_thread_load_context"

// ------------------------------------------------------------------------------
// hal_thread_load_context
// Load thread context
// r4 = address of sp of next thread to execute
// Note that this function is also the second half of hal_thread_switch_context
// and is simply dropped into from it.
	.global hal_thread_load_context
hal_thread_load_context:

    // disable interrupt(set PIE = 0)
    rdctl   r1,status
    andi    r1,r1,%lo(0xfffffffe)
    wrctl   status,r1

    ldw     r4,0(r4)
    addi    sp,r4, -SAVED_REGS_SIZE      // let sp = *NSP
    // restore all registers
    ldw     r0,  0(sp)
//    ldw     r1,  4(sp)                 // restore r1 at last
    ldw     r2,  8(sp)
    ldw     r3, 12(sp)
    ldw     r4, 16(sp)
    ldw     r5, 20(sp)
    ldw     r6, 24(sp)
    ldw     r7, 28(sp)
    ldw     r8, 32(sp)
    ldw     r9, 36(sp)
    ldw     r10, 40(sp)
    ldw     r11, 44(sp)
    ldw     r12, 48(sp)
    ldw     r13, 52(sp)
    ldw     r14, 56(sp)
    ldw     r15, 60(sp)
    ldw     r16, 64(sp)
    ldw     r17, 68(sp)
    ldw     r18, 72(sp)
    ldw     r19, 76(sp)
    ldw     r20, 80(sp)
    ldw     r21, 84(sp)
    ldw     r22, 88(sp)
    ldw     r23, 92(sp)
    ldw     r24, 96(sp)
    ldw     r25, 100(sp)    
    ldw     gp,  104(sp)
//    ldw     sp,  108(sp)        // restore sp at last      
    ldw     fp,  112(sp)        
    ldw     ea,  116(sp)
    ldw     ba,  120(sp)
    ldw     ra,  124(sp)
    // restore all ctrl registers
    ldw     r1,  128(sp)        // set PIE bit at last
    wrctl   status,r1
//    ldw     r1,  132(sp)
//    wrctl   estatus,r1
//    ldw     r1,  136(sp)
//    wrctl   bstatus,r1
//    ldw     r1,  140(sp)     // don't restore ienable register, keep as old value
//    wrctl   ienable,r1
//    ldw     r1,  144(sp)     // write to ipending reg is no effect
//    wrctl   ipending,r1
    
    // now , restore r1 & sp
    ldw     r1,4(sp)
    ldw     sp,108(sp)
    
    // return
    ret
    
#------------------------------------------------------------------------------
# HAL longjmp, setjmp implementations
# hal_setjmp saves all registers into buffer supplied in r4
	
	.globl	hal_setjmp
hal_setjmp:
    stw     r1, -4(sp)          // backup r1 first
    
    // save all registers to input buffer
    stw     r0, 0(r4)
    stw     r1, 4(r4)
    stw     r2, 8(r4)
    stw     r3, 12(r4)
    stw     r4, 16(r4)
    stw     r5, 20(r4)
    stw     r6, 24(r4)
    stw     r7, 28(r4)
    stw     r8, 32(r4)
    stw     r9, 36(r4)
    stw     r10, 40(r4)
    stw     r11, 44(r4)
    stw     r12, 48(r4)
    stw     r13, 52(r4)
    stw     r14, 56(r4)
    stw     r15, 60(r4)
    stw     r16, 64(r4)
    stw     r17, 68(r4)
    stw     r18, 72(r4)
    stw     r19, 76(r4)
    stw     r20, 80(r4)
    stw     r21, 84(r4)
    stw     r22, 88(r4)
    stw     r23, 92(r4)
    stw     r24, 96(r4)
    stw     r25, 100(r4)
    stw     gp, 104(r4)
    stw     sp, 108(r4)
    stw     fp, 112(r4)
    stw     ea, 116(r4)
    stw     ba, 120(r4)
    stw     ra, 124(r4)
    // save ctrl registers
    rdctl   r1, status
    stw     r1, 128(r4)
    rdctl   r1, estatus
    stw     r1, 132(r4)
    rdctl   r1, bstatus
    stw     r1, 136(r4)
    rdctl   r1, ienable
    stw     r1, 140(r4)
    rdctl   r1, ipending
    stw     r1, 144(r4)

    ldw     r1,-4(sp)
    mov     r2,r0               // set return value
    ret
// end of hal_setjmp()    
    
//---------------------------------
//hal_longjmp:
//     %r4 = input buffer pointer
//     %r5 = return value

	.globl	hal_longjmp
hal_longjmp:

    // disable interrupt(set PIE = 0)
    rdctl   r1,status
    andi    r1,r1,%lo(0xfffffffe)
    wrctl   status,r1

    mov     r2,r5                       // set return value first
    // restore all registers from input buffer
//    ldw     r1,  4(r4)                // resotre r1 at last
//    ldw     r2,  8(r4)                // don't need to restore r2, because r2 = return value
    ldw     r3, 12(r4)
//    ldw     r4, 16(r4)             // don't need to restore r4
    ldw     r5, 20(r4)
    ldw     r6, 24(r4)
    ldw     r7, 28(r4)
    ldw     r8, 32(r4)
    ldw     r9, 36(r4)
    ldw     r10, 40(r4)
    ldw     r11, 44(r4)
    ldw     r12, 48(r4)
    ldw     r13, 52(r4)
    ldw     r14, 56(r4)
    ldw     r15, 60(r4)
    ldw     r16, 64(r4)
    ldw     r17, 68(r4)
    ldw     r18, 72(r4)
    ldw     r19, 76(r4)
    ldw     r20, 80(r4)
    ldw     r21, 84(r4)
    ldw     r22, 88(r4)
    ldw     r23, 92(r4)
    ldw     r24, 96(r4)
    ldw     r25, 100(r4)    
    ldw     gp,  104(r4)  
//    ldw     sp,  108(r4)      // restore sp at last      
    ldw     fp,  112(r4)
    ldw     ea,  116(r4)
    ldw     ba,  120(r4)
    ldw     ra,  124(r4)
    // don't restore all ctrl registers
    ldw     r1,  128(r4)     // set PIE bit at last
    wrctl   status,r1
//    ldw     r1,  132(r4)
//    wrctl   estatus,r1
//    ldw     r1,  136(r4)
//    wrctl   bstatus,r1
//    ldw     r1,  140(r4)
//    wrctl   ienable,r1
//    ldw     r1,  144(r4)     // write to ipending reg is no effect
//    wrctl   ipending,r1
    
    // now , restore r1 & sp
    ldw     r1, 4(r4)
    ldw     sp, 108(r4)
    
    // return
    ret

//end of hal_longjmp()    
#------------------------------------------------------------------------------
# end of context.S
