//==========================================================================
//
//      *hal_misc.c
//
//      HAL miscellaneous functions
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Hao Tseng
// Contributors: 
// Date:         2003/11/13
// Purpose:      HAL miscellaneous functions
// Description:  This file contains miscellaneous functions provided by the
//               HAL.
//
//####DESCRIPTIONEND####
//
//========================================================================*/

#include <pkgconf/hal.h>

#include <cyg/infra/cyg_type.h>         // Base types

#include <cyg/hal/hal_arch.h>           // architectural definitions
#include <cyg/hal/hal_intr.h>           // Interrupt handling
#include <cyg/hal/hal_clock.h>
#include <cyg/hal/hal_diag.h>

#include <cyg/infra/cyg_type.h>
#include <cyg/infra/cyg_trac.h>         // tracing macros
#include <cyg/infra/cyg_ass.h>          // assertion macros
#include <cyg/infra/diag.h>

/*****************************************************************************
hal_default_exception_handler -- First level C exception handler

     The assembly default VSR  handler calls  this routine  to handler  the
exception.  When this routine returns, the  state is restored to the  state
pointed to by regs.

     We declare this  routine as  weak so  that other  handlers can  easily
become the default exception handler.

INPUT:

     vector: The exception vector number.

     regs: A pointer to the saved state.

OUTPUT:

RETURN VALUE:

     None

*****************************************************************************/

externC void hal_default_exception_handler(CYG_WORD vector, HAL_SavedRegisters *regs)
{

#ifdef CYGDBG_HAL_DEBUG_GDB_INCLUDE_STUBS
    externC void __handle_exception(void);
    externC HAL_SavedRegisters * _hal_registers;

    // Set the pointer to the registers of the current exception
    // context. At entry the GDB stub will expand the
    // HAL_SavedRegisters structure into a (bigger) register array.
    _hal_registers = regs;

    __handle_exception();

#elif defined(CYGFUN_HAL_COMMON_KERNEL_SUPPORT) && \
      defined(CYGPKG_HAL_EXCEPTIONS)

    // We should decode the vector and pass a more appropriate
    // value as the second argument. For now we simply pass a
    // pointer to the saved registers. We should also divert
    // breakpoint and other debug vectors into the debug stubs.

    cyg_hal_deliver_exception(vector, (CYG_ADDRWORD)regs);

#else

    CYG_FAIL("Exception!!!");

#endif

    return;
}

/*------------------------------------------------------------------------*/
/* hal_default_isr                                                        */
#ifdef CYGSEM_HAL_VIRTUAL_VECTOR_SUPPORT
externC cyg_uint32
hal_arch_default_isr(CYG_ADDRWORD vector, CYG_ADDRWORD data)
{
    CYG_TRACE1(true, "Interrupt: %d", vector);

    diag_printf("Spurious Interrupt!!! - vector: %d, data: %x\n", vector, 
                data);
    CYG_FAIL("Spurious Interrupt!!!");
    return 0;
}
#else

// interrupt service routine prototype
externC cyg_uint32 hal_default_isr(CYG_ADDRWORD vector, CYG_ADDRWORD data)
{
    CYG_TRACE1(true, "Interrupt: %d", vector);

#ifndef CYGDBG_HAL_DEBUG_GDB_INCLUDE_STUBS
#ifdef CYGDBG_HAL_DEBUG_GDB_CTRLC_SUPPORT
#ifdef CYGDBG_HAL_CTRLC_ISR
    // then see if it is an incoming character interrupt and break
    // into the stub ROM if the char is a ^C.
    if ( CYGDBG_HAL_CTRLC_ISR( vector, data ) )
        return 1; // interrupt handled
#endif
#endif
#endif

    diag_printf("Spurious Interrupt!!! - vector: %d, data: %x\n", vector, 
                data);
    CYG_FAIL("Spurious Interrupt!!!");
    return 0;
 
}

#endif


/*------------------------------------------------------------------------*/
/* data copy and bss zero functions                                       */

typedef void (CYG_SYM_ADDRESS)(void);

// All these must use this type of address to stop them being given relocations
// relative to $gp (i.e. assuming they would be in .sdata)
extern CYG_SYM_ADDRESS __ram_data_start;
extern CYG_SYM_ADDRESS __ram_data_end;
extern CYG_SYM_ADDRESS __rom_data_start;    

#ifdef CYG_HAL_STARTUP_ROM      
void hal_copy_data(void)
{
    cyg_uint32 *p = (cyg_uint32 *)&__ram_data_start;
    cyg_uint32 *q = (cyg_uint32 *)&__rom_data_start;
    cyg_uint32 *end = (cyg_uint32 *)&__ram_data_end;    

    while( p != end )
        *p++ = *q++;
}
#endif

void hal_zero_bss(void)
{
    extern CYG_SYM_ADDRESS __bss_start;
    extern CYG_SYM_ADDRESS __bss_end;

    cyg_uint32 zero = 0;
    cyg_uint32 *end = (cyg_uint32 *)&__bss_end;
    cyg_uint32 *p = (cyg_uint32 *)&__bss_start;


    while( p <= end )
        *p++ = zero;
    
}

/*------------------------------------------------------------------------*/
/* cyg_hal_invoke_constructors()                                          */

#ifdef CYGSEM_HAL_STOP_CONSTRUCTORS_ON_FLAG
cyg_bool hal_stop_constructors;
#endif

// use special hal_invoke_constructors() to avoid GCC's bug
typedef void (*pfunc) (void);
extern cyg_uint32 __CTOR_LIST__;
extern cyg_uint32 __CTOR_END__;

void
hal_invoke_constructors(void)
{
#ifdef CYGSEM_HAL_STOP_CONSTRUCTORS_ON_FLAG
    static cyg_uint32  *end,*list,func_addr;
    pfunc *p;
    
    list = &__CTOR_LIST__;    
    hal_stop_constructors = 0;    

    for (end = (&__CTOR_END__) -1; end >= list; end --) {
        func_addr = (cyg_uint32)(((cyg_uint32)(end) & 0xf0000000) | (*end >> 4));
        p = (pfunc *)(&func_addr);
        (*p) ();
        if (cyg_hal_stop_constructors) {
            end --;
            break;
        }
    }
    
#else
    cyg_uint32  *end,*list,func_addr;
    list = &__CTOR_LIST__;
    
    pfunc *p;

    for (end = (&__CTOR_END__) -1; end >= list; end --) {
        func_addr = (cyg_uint32)(((cyg_uint32)(end) & 0xf0000000) | (*end >> 4));
        p = (pfunc *)(&func_addr);
        (*p) ();
    }
#endif
} // hal_invoke_constructors()

/*------------------------------------------------------------------------*/
/* Determine the index of the ls bit of the supplied mask.                */

cyg_uint32 hal_lsbit_index(cyg_uint32 mask)
{
    cyg_uint32 n = mask;

    static const signed char tab[64] =
    { -1, 0, 1, 12, 2, 6, 0, 13, 3, 0, 7, 0, 0, 0, 0, 14, 10,
      4, 0, 0, 8, 0, 0, 25, 0, 0, 0, 0, 0, 21, 27 , 15, 31, 11,
      5, 0, 0, 0, 0, 0, 9, 0, 0, 24, 0, 0 , 20, 26, 30, 0, 0, 0,
      0, 23, 0, 19, 29, 0, 22, 18, 28, 17, 16, 0
    };

    n &= ~(n-1UL);
    n = (n<<16)-n;
    n = (n<<6)+n;
    n = (n<<4)+n;

    return tab[n>>26];
}

/*------------------------------------------------------------------------*/
/* Determine the index of the ms bit of the supplied mask.                */

cyg_uint32 hal_msbit_index(cyg_uint32 mask)
{
    cyg_uint32 x = mask;    
    cyg_uint32 w;

    /* Phase 1: make word with all ones from that one to the right */
    x |= x >> 16;
    x |= x >> 8;
    x |= x >> 4;
    x |= x >> 2;
    x |= x >> 1;

    /* Phase 2: calculate number of "1" bits in the word        */
    w = (x & 0x55555555) + ((x >> 1) & 0x55555555);
    w = (w & 0x33333333) + ((w >> 2) & 0x33333333);
    w = w + (w >> 4);
    w = (w & 0x000F000F) + ((w >> 8) & 0x000F000F);
    return (cyg_uint32)((w + (w >> 16)) & 0xFF) - 1;

}

/*------------------------------------------------------------------------*/
/* hal_idle_thread_action                                                 */
void hal_idle_thread_action(cyg_uint32 loop_count)
{
    // nothing to do
}


/*------------------------------------------------------------------------*/
/* Delay for some number of useconds.                                     */
void 
hal_delay_us(int us)
{ 
    cyg_uint32 val1, val2;
    long diff;
    long usticks;
    long ticks;

    // Calculate the number of counter register ticks per microsecond.
    usticks = CYGNUM_HAL_RTC_PERIOD / 10000;
    // Make sure that the value is not zero. This will only happen if the
    // CPU is running at < 2MHz.
    if( usticks == 0 ) usticks = 1;
    
    while( us > 0 )
    {
        int us1 = us;

        // Wait in bursts of less than 10000us to avoid any overflow
        // problems in the multiply.
        if( us1 > 10000 )
            us1 = 10000;

        us -= us1;

        ticks = usticks * us1;

        HAL_CLOCK_READ(&val1);
        while (ticks > 0) {
            HAL_CLOCK_READ(&val2);            
            diff = val2 - val1;
            if (diff < 0) diff += CYGNUM_HAL_RTC_PERIOD;
            ticks -= diff;
            val1 = val2;
        }
    }
}


/*------------------------------------------------------------------------*/
/* End of hal_misc.c                                                      */
