//========================================================================
//
//      nios2-stub.c
//
//      Helper functions for stub, generic to all NIOS II processors
//
//========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):     Hao
// Contributors:  
// Date:          2003/12/08
// Purpose:       
// Description:   Helper functions for stub, generic to all NIOS II processors
// Usage:         
//
//####DESCRIPTIONEND####
//
//========================================================================

#include <stddef.h>

#include <pkgconf/hal.h>

#ifdef CYGPKG_REDBOOT
#include <pkgconf/redboot.h>
#endif

#ifdef CYGDBG_HAL_DEBUG_GDB_INCLUDE_STUBS

#include <cyg/hal/hal_stub.h>

#define CYGARC_HAL_COMMON_EXPORT_CPU_MACROS
#include <cyg/hal/nios2_stub.h>

#include <cyg/hal/hal_arch.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/hal/hal_cache.h>

typedef cyg_uint32 t_inst;

/*--------------------------------------------------------------------*/
/* Given a trap value TRAP, return the corresponding signal. */

int __computeSignal (unsigned int trap_number)
{
  switch (trap_number)
    {
    case CYGNUM_HAL_VECTOR_TRAP:
    case CYGNUM_HAL_VECTOR_BREAK:    
        return SIGTRAP;
    default:
        return SIGINT;
    }
}

/* Return the trap number corresponding to the last-taken trap. */

int __get_trap_number (void)
{
//    cyg_uint32  reg_trap;
//    int trap_num;
//    
//    reg_trap = (get_register(REG_IENABLE)) & (get_register(REG_IPENDING));
//    for(trap_num=0;trap_num<32;trap_num++) {
//        if(reg_trap & 0x1)
//            return(trap_num);
//        reg_trap >>= 1;
//    }
//    return(trap_num);
    return(CYGNUM_HAL_VECTOR_BREAK);
}

#if defined(CYGSEM_REDBOOT_BSP_SYSCALLS)
int __is_bsp_syscall(void) 
{
    return 0;
}
#endif

/* Set the currently-saved pc register value to PC. This also updates NPC
   as needed. */

void set_pc (target_register_t pc)
{
  put_register (REG_PC, pc);
}


/*----------------------------------------------------------------------
 * Single-step support
 */

/* Saved instruction data for single step support.  */

static struct
{
  t_inst *targetAddr;
  t_inst savedInstr;
  t_inst *targetAddr2;
  t_inst savedInstr2;
  
} instrBuffer;


/* Set things up so that the next user resume will execute one instruction.
   This may be done by setting breakpoints or setting a single step flag
   in the saved user registers, for example. */

void __single_step (void)
{
    cyg_uint32 instruction;
    cyg_uint32  opcode;
    cyg_int16   imm16;
    cyg_uint32  imm26;
    cyg_uint32 pc = get_register(REG_PC);

    instrBuffer.targetAddr = (t_inst *)(pc + 4);              /* set default */
    instrBuffer.targetAddr2 = NULL;  
  
	/*
	 * Scan forward to see what the
	 * most appropriate location(s) for
	 * a breakpoint will be.
	 *
	 * The rules are:
     *  1.  I-TYPE: BEQ(0x26), BGE(0x0e), BGEU(0x2e), BLT(0x16), BLTU(0x36), BNE(0x1e)
     *  2.  I-TYPE: BR(0x06),
     *  3.  J-TYPE: CALL(0x0), 
     *  4.  R-TYPE: CALLR(0x3a,0x1d,0x0)(0xe83a), JMP(0x3a,0x0d,0x0)(0x683a),RET(0x3a,0x5,0x0)(0x283a)
     *  If *pc == (1), break at PC+4 and PC+4+IMM16
     *  If *pc == (2), break at PC+4+IMM16
     *  If *pc == (3), break at PC(31..28):IMM26*4
     *  If *pc == (4), break at RegA
     *  otherwise, just drop in the breakpoint
     */

	instruction = *(t_inst *)(pc);
    opcode = (instruction & 0x3f);
    if(opcode == 0x26 || opcode == 0x0e || opcode == 0x2e || opcode == 0x16 || opcode == 0x36 || opcode == 0x1e) {
        //case (1)
        imm16 = (cyg_int16)((instruction >> 6) & 0xffff);
        instrBuffer.targetAddr2 = (t_inst *)(pc + 4 + imm16); 
    }else if(opcode == 0x6) {
        //case (2)
        imm16 = (cyg_int16)((instruction >> 6) & 0xffff);        
        instrBuffer.targetAddr = (t_inst *)(pc + 4 + imm16);        
    }else if(opcode == 0x0) {
        //case(3)
        imm26 = instruction >> 6;
        instrBuffer.targetAddr = (t_inst *)((pc & 0xf0000000) | (imm26 << 2));
    }else if(opcode == 0x3a) {
        opcode = instruction & 0x1ffff;
        if(opcode == 0xe83a || opcode == 0x683a || opcode == 0x283a) {
            //case(4)
            instrBuffer.targetAddr = (t_inst *)get_register(instruction >> 27);
        }
    }
}


/* Clear the single-step state. */

void __clear_single_step (void)
{
//diag_printf("clear_ss ta %08x\n",instrBuffer.targetAddr);
  if (instrBuffer.targetAddr != NULL)
    {
      *instrBuffer.targetAddr = instrBuffer.savedInstr;
      HAL_NIOS2_INVALIDATE_DCACHE(instrBuffer.targetAddr);
      HAL_NIOS2_INVALIDATE_ICACHE(instrBuffer.targetAddr);
      HAL_NIOS2_INVALIDATE_PIPELINE();
      instrBuffer.targetAddr = NULL;
    }
  instrBuffer.savedInstr = NOP_INSTR;
  
  if (instrBuffer.targetAddr2 != NULL)
    {
      *instrBuffer.targetAddr2 = instrBuffer.savedInstr2;
      HAL_NIOS2_INVALIDATE_DCACHE(instrBuffer.targetAddr2);
      HAL_NIOS2_INVALIDATE_ICACHE(instrBuffer.targetAddr2);
      HAL_NIOS2_INVALIDATE_PIPELINE();
      instrBuffer.targetAddr2 = NULL;
    }
  instrBuffer.savedInstr2 = NOP_INSTR;

}


void __install_breakpoints ()
{
    
//diag_printf("install_bpt ta %08x\n",instrBuffer.targetAddr);
  if (instrBuffer.targetAddr != NULL)
    {
      instrBuffer.savedInstr = *instrBuffer.targetAddr;
      *instrBuffer.targetAddr = __break_opcode ();
      HAL_NIOS2_INVALIDATE_DCACHE(instrBuffer.targetAddr);
      HAL_NIOS2_INVALIDATE_ICACHE(instrBuffer.targetAddr);
      HAL_NIOS2_INVALIDATE_PIPELINE();
//diag_printf("ta %08x si %08x *ta %08x\n",
//            instrBuffer.targetAddr,instrBuffer.savedInstr,*instrBuffer.targetAddr);

    }
    
//diag_printf("install_bpt ta %08x\n",instrBuffer.targetAddr2);
  if (instrBuffer.targetAddr2 != NULL)
    {
      instrBuffer.savedInstr2 = *instrBuffer.targetAddr2;
      *instrBuffer.targetAddr2 = __break_opcode ();
      HAL_NIOS2_INVALIDATE_DCACHE(instrBuffer.targetAddr2);
      HAL_NIOS2_INVALIDATE_ICACHE(instrBuffer.targetAddr2);
      HAL_NIOS2_INVALIDATE_PIPELINE();
//diag_printf("ta %08x si %08x *ta %08x\n",
//            instrBuffer.targetAddr,instrBuffer.savedInstr2,*instrBuffer.targetAddr2);

    }
    

  /* Install the breakpoints in the breakpoint list */
  __install_breakpoint_list();
}

void __clear_breakpoints (void)
{
  __clear_breakpoint_list();
}


/* If the breakpoint we hit is in the breakpoint() instruction, return a
   non-zero value. */

int
__is_breakpoint_function ()
{
    return get_register (REG_PC) == (target_register_t)&_breakinst;
}


/* Skip the current instruction.  Since this is only called by the
   stub when the PC points to a breakpoint or trap instruction,
   we can safely just skip 2. */

void __skipinst (void)
{
    put_register (REG_PC, get_register (REG_PC) + 2);
}

  
    
#endif // CYGDBG_HAL_DEBUG_GDB_INCLUDE_STUBS
