//=============================================================================
//
//	vectors.S
//
//	NIOS II exception vectors
//
//=============================================================================
//###ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//###ECOSGPLCOPYRIGHTEND####
//=============================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):   Hao Tseng
// Contributors:    
// Date:        
// Purpose:	    NIOS II exception vectors
// Description:	This file defines the code placed into the exception
//              vectors. It also contains the first level default VSRs
//		        that save and restore state for both exceptions and
//		        interrupts.
//
//####DESCRIPTIONEND####
//
//=============================================================================

#include <pkgconf/system.h>	
#include <pkgconf/hal.h>

#ifdef CYGPKG_KERNEL
# include <pkgconf/kernel.h>	
#endif

#include  CYGBLD_HAL_PLATFORM_H
#include <cyg/hal/arch.inc>	
#include <cyg/hal/nios2_imacros.inc>
#include CYGHWR_MEMORY_LAYOUT_H

//-----------------------------------------------------------------------------
// (1) Reset vector

	// "ax": "a: section is allocatable" "x: section is executable"
	.section ".reset_vector","ax"
    .globl	reset_vector
    
    // we must do init icache in first 8 instructions
reset_vector:
    wrctl   status, r0       // Disabling interrupts
    
#ifdef CYGHWR_NIOS2_HAS_ICACHE    
    // Init icache    
    mov     r4, r0
    movhi   r5, %hi(CYGHWR_NIOS2_ICACHE_SIZE)
    ori     r5, r5, %lo(CYGHWR_NIOS2_ICACHE_SIZE)
_init_icache_loop:
    initi   r4
    // The LINE_SIZE always be 32
    addi    r4, r4, CYGHWR_NIOS2_ICACHE_LINE_SIZE       
    bltu    r4, r5, _init_icache_loop
    flushp
#endif

#ifdef CYGHWR_NIOS2_HAS_DCACHE
    //Init dcache
    mov     r4, r0
    movhi   r5, %hi(CYGHWR_NIOS2_DCACHE_SIZE)
    ori     r5, r5, %lo(CYGHWR_NIOS2_DCACHE_SIZE)
_init_dcache_loop:
    // This looped is unrolled by a factor of 8 to speed it up.
    // Further unrolling could be done as desired.
    initd   0*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    initd   1*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    initd   2*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    initd   3*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    initd   4*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    initd   5*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    initd   6*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    initd   7*CYGHWR_NIOS2_DCACHE_LINE_SIZE(r4)
    addi    r4, r4, CYGHWR_NIOS2_DCACHE_LINE_SIZE * 8
    bltu    r4, r5, _init_dcache_loop    
#endif
    
#ifdef CYG_HAL_STARTUP_ROMRAM
    //Copy ROM program to RAM, then jump to RAM
    .extern	__ram_data_end    
    movhi   r4, %hi(CYGMEM_REGION_rom)              // r4 = rom begin address
    ori     r4, r4, %lo(CYGMEM_REGION_rom)
    movhi   r5, %hi(CYGMEM_REGION_ram)              // r5 = ram begin address
    ori     r5, r5, %lo(CYGMEM_REGION_ram)
    movhi   r3, %hi(__ram_data_end)                 // r3 = the end of code address in ram
    ori     r3, r3, %lo(__ram_data_end)
_romram_loop:    
    ldwio   r1, 0(r4)
    stwio   r1, 0(r5)
    ldwio   r1, 4(r4)
    stwio   r1, 4(r5)
    ldwio   r1, 8(r4)
    stwio   r1, 8(r5)
    ldwio   r1, 12(r4)
    stwio   r1, 12(r5)
    addi    r4,r4,16
    addi    r5,r5,16
    bltu    r5,r3, _romram_loop
    
    movhi   r4, %hi(_ram_begin)
    ori     r4,r4,%lo(_ram_begin)
    jmp     r4
_ram_begin:    
#endif

    wrctl   ienable,r0
    
    //(2) Set default exception handler    
    movhi   r2,%hi(_jump_to_exception_handler)
    ori     r2,r2,%lo(_jump_to_exception_handler)
    movhi   r3,%hi(CYGHWR_NIOS2_EXCEPTION_ADDRESS)
    ori     r3,r3,%lo(CYGHWR_NIOS2_EXCEPTION_ADDRESS)
    
    // copy three instructions to CYGHWR_NIOS2_EXCEPTION_ADDRESS
    ldw     r4,0(r2)
    stw     r4,0(r3)
    ldw     r4,4(r2)
    stw     r4,4(r3)
    ldw     r4,8(r2)
    stw     r4,8(r3)
    
    // flush code & data cache of CYGHWR_NIOS2_EXCEPTION_ADDRESS
#ifdef CYGHWR_NIOS2_HAS_DCACHE     
    flushd  0(r3)
    flushd  4(r3)
    flushd  8(r3)    
#endif    
#ifdef CYGHWR_NIOS2_HAS_ICACHE
    flushi  r3
    addi    r3,r3,4
    flushi  r3
    addi    r3,r3,4
    flushi  r3
    flushp
#endif    
    
#ifndef CYGSEM_HAL_USE_ROM_MONITOR    
    // Work as ROM Monitor
    //(3) Set default break handler    
    movhi   r2,%hi(_jump_to_break_handler)
    ori     r2,r2,%lo(_jump_to_break_handler)
    movhi   r3,%hi(CYGHWR_NIOS2_BREAK_ADDRESS)
    ori     r3,r3,%lo(CYGHWR_NIOS2_BREAK_ADDRESS)
    
    // copy three instructions to CYGHWR_NIOS2_BREAK_ADDRESS
    ldw     r4,0(r2)
    stw     r4,0(r3)
    ldw     r4,4(r2)
    stw     r4,4(r3)
    ldw     r4,8(r2)
    stw     r4,8(r3)
    
    // flush code & data cache of CYGHWR_NIOS2_BREAK_ADDRESS
#ifdef CYGHWR_NIOS2_HAS_DCACHE    
    flushd  0(r3)
    flushd  4(r3)
    flushd  8(r3)    
#endif
#ifdef CYGHWR_NIOS2_HAS_ICACHE
    flushi  r3
    addi    r3,r3,4
    flushi  r3
    addi    r3,r3,4
    flushi  r3
    flushp
#endif    
#endif  //#ifndef CYGSEM_HAL_USE_ROM_MONITOR
    
    //(4) reset platfrom hardware
    .extern	hal_platform_reset
    movhi   r7, %hi(hal_platform_reset)
    ori     r7, r7, %lo(hal_platform_reset)
    callr   r7

    //(5) setup stack pointer and frame pointer and global pointer
    movhi	sp, %hi(cyg_interrupt_stack)
    ori	    sp, sp,%lo(cyg_interrupt_stack)
    mov     fp, sp
    .extern _gp
    movhi	gp, %hi(_gp)
    ori	    gp, gp, %lo(_gp)

#ifdef CYG_HAL_STARTUP_ROM
    //(6) Copy data from ROM to RAM if the program is start from ROM rather than RAM
    movhi   r7, %hi(hal_copy_data)
    ori     r7, r7, %lo(hal_copy_data)
    callr   r7
#endif

    //(7) Zero BSS
	.extern	hal_zero_bss
    movhi   r7, %hi(hal_zero_bss)
    ori     r7, r7, %lo(hal_zero_bss)
    callr   r7

    //(8) platfrom flushialize (flushial hardware)
	.extern	hal_platform_init
    movhi   r7, %hi(hal_platform_init)
    ori     r7, r7, %lo(hal_platform_init)
    callr   r7

//    //(9) Call constructors 
//	.extern	_do_ctors
//    movhi   r7, %hi(_do_ctors)
//    ori     r7, r7, %lo(_do_ctors)
//    callr   r7
    
    //(9) Call constructors 
	.extern	hal_invoke_constructors
    movhi   r7, %hi(hal_invoke_constructors)
    ori     r7, r7, %lo(hal_invoke_constructors)
    callr   r7
    
#ifdef CYGDBG_HAL_DEBUG_GDB_INCLUDE_STUBS
    //(10) init GDB_STUB
	.extern	initialize_stub
    movhi   r7, %hi(initialize_stub)
    ori     r7, r7, %lo(initialize_stub)
    callr   r7
#endif

#if defined(CYGDBG_HAL_DEBUG_GDB_CTRLC_SUPPORT) \
    || defined(CYGDBG_HAL_DEBUG_GDB_BREAK_SUPPORT)
    //(11) init GDB_STUB ctrl_c support
    .extern hal_ctrlc_isr_init
    movhi   r7, %hi(hal_ctrlc_isr_init)
    ori     r7, r7, %lo(hal_ctrlc_isr_init)
    callr   r7
#endif

    //(12) Call cyg_start of the ecos
	.extern	cyg_start
    movhi   r7, %hi(cyg_start)
    ori     r7, r7, %lo(cyg_start)
    callr   r7
    
    // Loop forever if return
_LoopForever:
    br      _LoopForever
    
//-----------------------------------------------------------------------------
// (2) Execption & Break
//
_jump_to_exception_handler:
    movhi   r24,%hi(InterruptVSRHandler)
    ori     r24,r24,%lo(InterruptVSRHandler)
    jmp     r24

_jump_to_break_handler:
    movhi   r25,%hi(BreakHandler)
    ori     r25,r25,%lo(BreakHandler)
    jmp     r25

BreakHandler:
	// PIE cleared here
    //======================================================
    //(1)Save CPU State
    // We create a space in interruptee's stack for HAL_SavedRegisters, return address
    //      (thread'sp):  xxxx   
    //                    yyyy  ---    return address(PC)
    //                     |
    //                     |     HAL_SavedRegisters (SAVED_REGS_SIZE)
    //                     |     
    //             NSP--> zzzz  ---    r[0]
    //
    // let sp = NSP
    addi    sp,sp, -SAVED_REGS_SIZE
    stw     r0, 0(sp)
    stw     r1, 4(sp)
    stw     r2, 8(sp)
    stw     r3, 12(sp)
    stw     r4, 16(sp)
    stw     r5, 20(sp)
    stw     r6, 24(sp)
    stw     r7, 28(sp)
    stw     r8, 32(sp)
    stw     r9, 36(sp)
    stw     r10, 40(sp)
    stw     r11, 44(sp)
    stw     r12, 48(sp)
    stw     r13, 52(sp)
    stw     r14, 56(sp)
    stw     r15, 60(sp)
    stw     r16, 64(sp)
    stw     r17, 68(sp)
    stw     r18, 72(sp)
    stw     r19, 76(sp)
    stw     r20, 80(sp)
    stw     r21, 84(sp)
    stw     r22, 88(sp)
    stw     r23, 92(sp)
    stw     r24, 96(sp)
    stw     r25, 100(sp)
    stw     gp, 104(sp)
    addi    r1,sp,SAVED_REGS_SIZE
    stw     r1, 108(sp)
    stw     fp, 112(sp)
    stw     ea, 116(sp)
    stw     ba, 120(sp)
    stw     ra, 124(sp)
    // save ctrl registers
    rdctl   r1, status
    stw     r1, 128(sp)
    rdctl   r1, estatus
    stw     r1, 132(sp)
    rdctl   r1, bstatus
    stw     r1, 136(sp)
    rdctl   r1, ienable
    stw     r1, 140(sp)
    rdctl   r1, ipending
    stw     r1, 144(sp)
    // save return address
    addi    r1, ba, -4
    stw     r1, 148(sp)

    .extern _gp
    movhi	gp, %hi(_gp)                        // before ISR handler(), restore original gp value. (for GDB)
    ori	    gp, gp, %lo(_gp)
    
    //======================================================
    //(4)Switch to an interrupt stack or using thread's stack
    mov     r1,sp       // let r1= NSP
#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
    //======================================================
    // Using interrupt stack

	// get interrupt stack address into %r4 (top -- High)
    movhi   r4, %hi(cyg_interrupt_stack)
    ori     r4, r4, %lo(cyg_interrupt_stack)
    
	// get the address of cyg_interrupt_stack_base into r5 
	// (bottom -- low)
    movhi   r5, %hi(cyg_interrupt_stack_base)
    ori     r5, r5, %lo(cyg_interrupt_stack_base)

    bltu    sp, r5, _BISR1      // below stack base?, if yes, switch new stack
    bltu    sp, r4, _BISR2      // below top? then no switch needed

_BISR1:	// switch to the interrupt stack
    mov     sp, r4
    
_BISR2:	
	  // continue as before, if already on the interrupt stack
#endif

    addi    sp, sp, -16         // reserve 4 words for variables
    mov     fp, sp              // fp[0],fp[1], fp[2], fp[3]
    stw     r1, 0(fp)           // fp[0] = NSP

    mov     r4,r0               // let r4 = 0 (break vector)    
    slli    r1,r4,2                     //r1 = r4 << 2
    .extern hal_vsr_table    
    movhi   r2, %hi(hal_vsr_table) 
    ori     r2, r2, %lo(hal_vsr_table)
    add     r2,r2,r1                    // r2 = &hal_vsr_table[r1]
    ldw     r7,0(r2)                    // r7 = hal_vsr_table[exception_num]
    
    // call ISR()
    addi    r4,r4,CYGNUM_HAL_EXCEPTION_MIN      // r4 = exception_num
    ldw     r5, 0(fp)                           // r5 = NSP
    callr   r7                                  // call ExceptionHandler(exception_num, NSP)
    mov     fp,sp

    // return process
    ldw     sp,  0(fp)                  // let sp = NSP
    // restore all registers
    ldw     r0,  0(sp)
//    ldw     r1,  4(sp)                // restore r1 later
    ldw     r2,  8(sp)
    ldw     r3, 12(sp)
    ldw     r4, 16(sp)
    ldw     r5, 20(sp)
    ldw     r6, 24(sp)
    ldw     r7, 28(sp)
    ldw     r8, 32(sp)
    ldw     r9, 36(sp)
    ldw     r10, 40(sp)
    ldw     r11, 44(sp)
    ldw     r12, 48(sp)
    ldw     r13, 52(sp)
    ldw     r14, 56(sp)
    ldw     r15, 60(sp)
    ldw     r16, 64(sp)
    ldw     r17, 68(sp)
    ldw     r18, 72(sp)
    ldw     r19, 76(sp)
    ldw     r20, 80(sp)
    ldw     r21, 84(sp)
    ldw     r22, 88(sp)
    ldw     r23, 92(sp)
    ldw     r24, 96(sp)
    ldw     r25, 100(sp)    
    ldw     gp,  104(sp)
//    ldw     sp,  108(sp)      // restore sp at last
    ldw     fp,  112(sp)
    ldw     ea,  116(sp)
    ldw     ba,  148(sp)        // sp[148] = pc = return address
    ldw     ra,  124(sp)
    // restore all ctrl registers
//    ldw     r1,  128(sp)      // don't need to restore status reg
//    wrctl   status,r1
    ldw     r1,  132(sp)
    wrctl   estatus,r1
    ldw     r1,  136(sp)
    wrctl   bstatus,r1
    ldw     r1,  140(sp)
    wrctl   ienable,r1
//    ldw     r1,  144(sp)     // write to ipending reg is no effect
//    wrctl   ipending,r1
    
    // now , restore r1 & sp
    ldw     r1,4(sp)
    ldw     sp,108(sp)
    
    // return
    bret
    
// (2)-1 Null interrupt handler
//	
NullInterruptHandler:
    eret

//
//(2)-4 Interrupt VSR handler
InterruptVSRHandler:

	// PIE cleared here

    //======================================================
    //(1)Save CPU State
    // We create a space in interruptee's stack for HAL_SavedRegisters, return address
    //      (thread'sp):  xxxx   
    //                    yyyy  ---    return address(PC)
    //                     |
    //                     |     HAL_SavedRegisters (SAVED_REGS_SIZE)
    //                     |     
    //             NSP--> zzzz  ---    r[0]
    //
    // let sp = NSP
    addi    sp,sp, -SAVED_REGS_SIZE
    stw     r0, 0(sp)
    stw     r1, 4(sp)
    stw     r2, 8(sp)
    stw     r3, 12(sp)
    stw     r4, 16(sp)
    stw     r5, 20(sp)
    stw     r6, 24(sp)
    stw     r7, 28(sp)
    stw     r8, 32(sp)
    stw     r9, 36(sp)
    stw     r10, 40(sp)
    stw     r11, 44(sp)
    stw     r12, 48(sp)
    stw     r13, 52(sp)
    stw     r14, 56(sp)
    stw     r15, 60(sp)
    stw     r16, 64(sp)
    stw     r17, 68(sp)
    stw     r18, 72(sp)
    stw     r19, 76(sp)
    stw     r20, 80(sp)
    stw     r21, 84(sp)
    stw     r22, 88(sp)
    stw     r23, 92(sp)
    stw     r24, 96(sp)
    stw     r25, 100(sp)
    stw     gp, 104(sp)
    addi    r1,sp,SAVED_REGS_SIZE
    stw     r1, 108(sp)
    stw     fp, 112(sp)
    stw     ea, 116(sp)
    stw     ba, 120(sp)
    stw     ra, 124(sp)
    // save ctrl registers
    rdctl   r1, status
    stw     r1, 128(sp)
    rdctl   r1, estatus
    stw     r1, 132(sp)
    rdctl   r1, bstatus
    stw     r1, 136(sp)
    rdctl   r1, ienable
    stw     r1, 140(sp)
    rdctl   r1, ipending
    stw     r1, 144(sp)
    // save return address
    addi    r1, ea, -4
    stw     r1, 148(sp)
        
    .extern _gp
    movhi	gp, %hi(_gp)                        // before ISR handler(), restore original gp value. (for GDB)
    ori	    gp, gp, %lo(_gp)
    
    //======================================================
    //(4)Switch to an interrupt stack or using thread's stack
    mov     r1,sp       // let r1= NSP
#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
    //======================================================
    // Using interrupt stack

	// get interrupt stack address into %r4 (top -- High)
    movhi   r4, %hi(cyg_interrupt_stack)
    ori     r4, r4, %lo(cyg_interrupt_stack)
    
	// get the address of cyg_interrupt_stack_base into r5 
	// (bottom -- low)
    movhi   r5, %hi(cyg_interrupt_stack_base)
    ori     r5, r5, %lo(cyg_interrupt_stack_base)

    bltu    sp, r5, _EISR1      // below stack base?, if yes, switch new stack
    bltu    sp, r4, _EISR2      // below top? then no switch needed

_EISR1:	// switch to the interrupt stack
    mov     sp, r4
    
_EISR2:	
	  // continue as before, if already on the interrupt stack
#endif

    addi    sp, sp, -16         // reserve 4 words for variables
    mov     fp, sp              // fp[0],fp[1], fp[2], fp[3]
    stw     r1, 0(fp)           // fp[0] = NSP

    //======================================================
    //(5) Decide interrupt type (Exception or Interrupt)
    rdctl   r24, estatus
    andi    r24, r24, 1
    beq     r24, r0, _type_exception
    rdctl   r24, ipending
    bne     r24, r0, _type_interrupt
    
// Exception or software interrupt    
_type_exception:
    mov     r4,r0                   // assume r4(exception num) = 1 (software trap)                       
    ldw     r1, -4(ea)
    movhi   r2,%hi(0x003d683a)      
    ori     r2,r2,%lo(0x003d683a)
    beq     r1,r2,_call_exception_isr     // check the instruction is 'trap' or not
    addi    r4,r4,1                 // set r4(exception num) = 2 (exception happen)
    
_call_exception_isr:
    slli    r1,r4,2                     //r1 = r4 << 2
    .extern hal_vsr_table    
    movhi   r2, %hi(hal_vsr_table) 
    ori     r2, r2, %lo(hal_vsr_table)
    add     r2,r2,r1                    // r2 = &hal_vsr_table[r1]
    ldw     r7,0(r2)                    // r7 = hal_vsr_table[exception_num]
    
    // call ISR()
    addi    r4,r4,CYGNUM_HAL_EXCEPTION_MIN      // r4 = exception_num
    ldw     r5, 0(fp)                           // r5 = NSP
    callr   r7                                  // call ExceptionHandler(exception_num, NSP)
    mov     fp,sp
    br      _return_proc   
    // end of exception handler

// Interrupt 
_type_interrupt:
    //======================================================
    //(3)Increment the kernel scheduler lock
    //    
#ifdef CYGFUN_HAL_COMMON_KERNEL_SUPPORT
	// Lock the scheduler
	// load scheduler counter into r2    
    // PS: the cyg_scheduler_sched_lock is define as volatile variables.
    //     So, must use ldwio & stwio for it
	.extern	cyg_scheduler_sched_lock
    movhi   r1, %hi(cyg_scheduler_sched_lock)
    ori     r1, r1, %lo(cyg_scheduler_sched_lock)
    ldwio   r2, 0(r1)
	// lock scheduler    
    addi    r2,r2,1
    // save back
    stwio   r2, 0(r1)
#endif
    //======================================================
    //(5) Decode the actual interrupt number
    rdctl   r1, ienable
    rdctl   r2, ipending
    and     r1,r1,r2            // r1 = ienable & ipending
    mov     r4,r0               // let r4 = 0, (used for calculate irq number)
    addi    r3,r0,1             // let r3 = mask bit
_int_loop:                      // check irq loop
    and     r2,r1,r3
    bne     r2,r0,_get_irq_num
    slli    r3,r3,1
    addi    r4,r4,1    
    br      _int_loop
_get_irq_num:
    // mask the interrupt bit
//    nor     r3,r3,r0            // r3 = ~ r3
//    rdctl   r1,ienable
//    and     r1,r1,r3
//    wrctl   ienable, r1
    
    // now, r4 = isr_vector_num
    stw     r4, 4(fp)           // fp[1] = irq_num
    //======================================================
    //(2)Platform interrupt pre_process
    //    
    .extern hal_platform_intr_pre_process
    movhi   r7, %hi(hal_platform_intr_pre_process)
    ori     r7, r7, %lo(hal_platform_intr_pre_process)
    callr   r7
    mov     fp,sp

    ldw     r4, 4(fp)           // restore irq_num
    addi    r4,r4,-CYGNUM_HAL_ISR_MIN    // vector_num = irq_num - CYGNUM_HAL_ISR_MIN
    //======================================================
    //(6)Get ISR pointer and it's data pointer
    // Get hal_interrupt_data[isr_vector_num]
    slli    r1,r4,2                     //r1 = r4 << 2
    .extern hal_interrupt_data    
    movhi   r2, %hi(hal_interrupt_data) 
    ori     r2, r2, %lo(hal_interrupt_data)
    add     r2,r2,r1                    // r2 = &hal_interrupt_data[r1]
    ldw     r5,0(r2)                    // r5 = hal_interrupt_data[isr_vector_num]
    // Get *hal_interrupt_handlers[isr_vector_num]
    .extern hal_interrupt_handlers
    movhi   r2, %hi(hal_interrupt_handlers) 
    ori     r2, r2, %lo(hal_interrupt_handlers)
    add     r2,r2,r1                    // r2 = &hal_interrupt_handlers[r1]
    ldw     r7,0(r2)                    // r7 = *hal_interrupt_handlers[isr_vector_num]    
    
    //======================================================
    //(7) Re-enable interrupt to permit nesting
//    rdctl   r1, status
//    ori     r1,r1,1                     // set PIE = 1
//    wrctl   status,r1
    
    //======================================================
    //(8)Call the ISR
    callr   r7                          // ISR(vector, hal_interrupt_data[ vector ])
    mov     fp,sp    
    mov     r4,r2                       // let r4 = return value
    ldw     r5, 4(fp)                   // let r5 = irq_num
    ldw     r6, 0(fp)                   // let r6 = NSP
    
#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
	// If we are returning from the last nested interrupt, move back
	// to the thread stack. interrupt_end() must be called on the
	// thread stack since it potentially causes a context switch.
    
    // disable interrupt(set PIE = 0)
    rdctl   r1,status
    andi    r1,r1,%lo(0xfffffffe)
    wrctl   status,r1
    
    mov     sp, r6              // switch sp back to NSP
    addi    sp, sp, -16         // reserve 4 words for variables
    mov     fp, sp              // fp[0],fp[1], fp[2], fp[3]
    stw     r6, 0(fp)           // fp[0] = NSP
    stw     r5, 4(fp)           // fp[1] = irq_num
    
    // enable interrupt (set PIE = 1)
    rdctl   r1, status
    ori     r1,r1,1                     // set PIE = 1
    wrctl   status,r1
    
#endif    

#ifdef CYGFUN_HAL_COMMON_KERNEL_SUPPORT
    //======================================================
    //(9)Call interrupt_end(r, cyg_hal_interrupt_objects[vector], regs)
	// - r is the return value from the ISR
	// - regs points to saved CPU context
    .extern hal_interrupt_objects
    slli    r1,r5,2                     // r1 = r5(irq_number) << 2
    movhi   r5, %hi(hal_interrupt_objects) 
    ori     r5, r5, %lo(hal_interrupt_objects)
    add     r5,r5,r1                    // r5 = &hal_interrupt_objects[r1]
    ldw     r5, 0(r5)                   // r5 = hal_interrupt_objects[vectors]
    // now r4=ISR return value
    //     r5 = hal_interrupt_objects[vectors]
    //     r6 = NSP    
  	.extern	interrupt_end
    movhi   r7, %hi(interrupt_end)
    ori     r7, r7, %lo(interrupt_end)
    callr   r7                           // Call interrupt_end()
    mov     fp,sp    
#endif    

    // disable interrupt(set PIE = 0)
    rdctl   r1,status
    andi    r1,r1,%lo(0xfffffffe)
    wrctl   status,r1

    //======================================================
    //(10)return process
_return_proc:    
    ldw     sp,  0(fp)                  // let sp = NSP
    
    // restore all registers
    ldw     r0,  0(sp)
//    ldw     r1,  4(sp)                // restore r1 later
    ldw     r2,  8(sp)
    ldw     r3, 12(sp)
    ldw     r4, 16(sp)
    ldw     r5, 20(sp)
    ldw     r6, 24(sp)
    ldw     r7, 28(sp)
    ldw     r8, 32(sp)
    ldw     r9, 36(sp)
    ldw     r10, 40(sp)
    ldw     r11, 44(sp)
    ldw     r12, 48(sp)
    ldw     r13, 52(sp)
    ldw     r14, 56(sp)
    ldw     r15, 60(sp)
    ldw     r16, 64(sp)
    ldw     r17, 68(sp)
    ldw     r18, 72(sp)
    ldw     r19, 76(sp)
    ldw     r20, 80(sp)
    ldw     r21, 84(sp)
    ldw     r22, 88(sp)
    ldw     r23, 92(sp)
    ldw     r24, 96(sp)
    ldw     r25, 100(sp)    
    ldw     gp,  104(sp)
//    ldw     sp,  108(sp)      // restore sp at last
    ldw     fp,  112(sp)
    ldw     ea,  148(sp)        // sp[148] = pc = return address
    ldw     ba,  120(sp)
    ldw     ra,  124(sp)
    // restore all ctrl registers
//    ldw     r1,  128(sp)      // don't need to restore status reg
//    wrctl   status,r1
    ldw     r1,  132(sp)
    wrctl   estatus,r1
    ldw     r1,  136(sp)
    wrctl   bstatus,r1
//    ldw     r1,  140(sp)     // don't restore ienable, keep the newest status
//    wrctl   ienable,r1
//    ldw     r1,  144(sp)     // write to ipending reg is no effect
//    wrctl   ipending,r1
    
    // now , restore r1 & sp
    ldw     r1,4(sp)
    ldw     sp,108(sp)
    
    // return
    eret
//InterruptVSRHandler finish.


//
//(2)-5 hal_interrupt_stack_call_pending_DSRs

// only applys to the interrupt stack used
#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK

// hal_interrupt_stack_call_pending_DSRs will **not** be called in nested interrupt coz unlock_inner( ) will not called
// which will in turns call Cyg_Interrupt::call_pending_DSRs( ) -->  HAL_INTERRUPT_STACK_CALL_PENDING_DSRS( ) 
// --> hal_interrupt_stack_call_pending_DSRs( ) (this routine)

// This routine can **only** be called from a thread stack
// It switches to the interrupt stack then calls back to the
// kernel to execute DSRs.

	.globl hal_interrupt_stack_call_pending_DSRs
hal_interrupt_stack_call_pending_DSRs:
    // disable interrupt(set PIE = 0)
    rdctl   r1,status
    andi    r1,r1,%lo(0xfffffffe)
    wrctl   status,r1

    movhi       r1,%hi(cyg_interrupt_stack)
    ori         r1,r1,%lo(cyg_interrupt_stack)
    stw         sp, -4(r1)
    mov         sp, r1
    addi        sp,sp,-8
    stw         ra, 0(sp)
    
    // enable interrupt (set PIE = 1)
    rdctl   r1, status
    ori     r1,r1,1                     // set PIE = 1
    wrctl   status,r1
    
	// call cyg_interrupt_call_pending_DSRs
	.extern cyg_interrupt_call_pending_DSRs
    movhi       r1,%hi(cyg_interrupt_call_pending_DSRs)
    ori         r1,r1,%lo(cyg_interrupt_call_pending_DSRs)
    callr       r1
    
    ldw         ra, 0(sp)
    ldw         sp, 4(sp)
    ret

#endif // CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK



//-----------------------------------------------------------------------------
// (3) Define startup stack & interrupt stack
//
// cyg_interrupt_stack			|	"*high* memory address"
//								|	
//								|
//								|
//								|
// cyg_interrupt_stack_base     V	"*low* memory address"

// Note: stack grows downward
// we define the startup stack


#ifndef CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE
#define CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE 4096
#endif

   	.section ".bss"
	.balign 4
	.globl cyg_interrupt_stack_base
cyg_interrupt_stack_base:
	
	.rept CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE
	.word 0
	.endr

	.balign 4
	.globl cyg_interrupt_stack
cyg_interrupt_stack:
	.rept 64
	.word 0
	.endr


    
//-----------------------------------------------------------------------------
// Interrupt Table
//
// These tables contain the isr, data and object pointers used to deliver
// interrupts to user code.
// hal_interrupt_level contains the interrupt level set by 
// HAL_INTERRUPT_CONFIGURE().
// This is a default set that provide support only for the 6 external
// interrupts in the status/cause registers. Platforms or boards are expected
// to define their own versions of these if they have their own interrupt mappings.

	.data
//	.section ".data"
	.balign 4    
    
	.extern hal_default_isr
    
	.globl	hal_interrupt_handlers
hal_interrupt_handlers:
    .rept   CYGNUM_HAL_ISR_COUNT
	.long	hal_default_isr
    .endr

	.globl	hal_interrupt_data
hal_interrupt_data:
	.rept	CYGNUM_HAL_ISR_COUNT
	.long	0
	.endr

	.globl	hal_interrupt_objects
hal_interrupt_objects:
	.rept	CYGNUM_HAL_ISR_COUNT
	.long	0
	.endr

//-----------------------------------------------------------------------------
// VSR Table
//
	.data
	.balign 4    
    
	.extern hal_default_exception_handler
    
	.globl	hal_vsr_table
hal_vsr_table:
    .rept   CYGNUM_HAL_EXCEPTION_COUNT
	.long	hal_default_exception_handler
    .endr
    
//   	.section ".ctors_begin"
//	.balign 4
//	.globl _do_ctors
//_do_ctors:
//    addi    sp,sp,-8
//    stw     ra,4(sp)
//    
//   	.section ".ctors_end"
//	.balign 4
//    ldw     ra,4(sp)
//    addi    sp,sp,8
//    ret
//    call	0

//-----------------------------------------------------------------------------
// end of vectors.S


